import pytest
import scipy
import numpy as np
import numpy.testing as npt
import ukat.utils.arraystats as ast


def calc_stats(x):
    """
    Gold standard (gs) implementation of calculation of statistical measures.

    Parameters
    ----------
    x : np.ndarray with 1 dimension

    Notes
    -----
    This is defined at the module level as it is used by both test Classes.

    """
    n = len(x)
    mean = np.mean(x)
    median = np.median(x)
    minimum = np.min(x)
    maximum = np.max(x)
    std = np.std(x)
    if mean == 0:
        cv = np.nan
    else:
        cv = std/mean
    skewness = scipy.stats.skew(x, bias=True)
    kurtosis = scipy.stats.kurtosis(x, fisher=True, bias=True)
    entropy = scipy.stats.entropy(x)

    return [n, mean, median, minimum, maximum, std, cv, skewness, kurtosis,
            entropy]


class TestFlatStats:

    def test_array_with_one_element(self):
        # Note `gs`: gold standard

        # Int
        x1 = np.array([5])
        gs_1 = calc_stats(x1)
        stats_1 = self.flatstats_to_array(ast.FlatStats(x1).calculate())
        npt.assert_allclose(stats_1, gs_1, rtol=1e-6)

        # Float
        x2 = np.array([5.3])
        gs_2 = calc_stats(x2)
        stats_2 = self.flatstats_to_array(ast.FlatStats(x2).calculate())
        npt.assert_allclose(stats_2, gs_2, rtol=1e-6)

        # Negative number
        x3 = np.array([-5.3])
        gs_3 = calc_stats(x3)
        stats_3 = self.flatstats_to_array(ast.FlatStats(x3).calculate())
        npt.assert_allclose(stats_3, gs_3, rtol=1e-6)

        # Zero
        x4 = np.array([0])
        gs_4 = calc_stats(x4)
        stats_4 = self.flatstats_to_array(ast.FlatStats(x4).calculate())
        npt.assert_allclose(stats_4, gs_4, rtol=1e-6, equal_nan=True)

        # Nan
        x5 = np.array([np.nan])
        gs_5 = calc_stats(x5)
        stats_5 = self.flatstats_to_array(ast.FlatStats(x5).calculate())
        npt.assert_allclose(stats_5, gs_5, rtol=1e-6, equal_nan=True)

    def test_array_with_multiple_elements(self):
        # Ints
        x1 = np.array([5, 9, 7, 12, 19])
        gs_1 = calc_stats(x1)
        stats_1 = self.flatstats_to_array(ast.FlatStats(x1).calculate())
        npt.assert_allclose(stats_1, gs_1, rtol=1e-6)

        # Float
        x2 = np.array([5, 9, np.sqrt(7), 12, 19])
        gs_2 = calc_stats(x2)
        stats_2 = self.flatstats_to_array(ast.FlatStats(x2).calculate())
        npt.assert_allclose(stats_2, gs_2, rtol=1e-6)

        # Negative numbers
        x3 = np.array([5, 9, -np.sqrt(7), -12, 19])
        gs_3 = calc_stats(x3)
        stats_3 = self.flatstats_to_array(ast.FlatStats(x3).calculate())
        npt.assert_allclose(stats_3, gs_3, rtol=1e-6)

        # Zero
        x4 = np.array([5, 0, np.sqrt(7), -12, 19])
        gs_4 = calc_stats(x4)
        stats_4 = self.flatstats_to_array(ast.FlatStats(x4).calculate())
        npt.assert_allclose(stats_4, gs_4, rtol=1e-6, equal_nan=True)

        # Nan
        x5 = np.array([5, 0, np.sqrt(7), -12, np.nan])
        gs_5 = calc_stats(x5)
        stats_5 = self.flatstats_to_array(ast.FlatStats(x5).calculate())
        npt.assert_allclose(stats_5, gs_5, rtol=1e-6, equal_nan=True)

    def test_empty_array(self):
        expected_stats = np.array([0,        # n
                                   np.nan,   # mean
                                   np.nan,   # median
                                   np.nan,   # min
                                   np.nan,   # max
                                   np.nan,   # std
                                   np.nan,   # cv
                                   np.nan,   # skewness
                                   np.nan,   # kurtosis
                                   np.nan])  # entropy

        x1 = np.array([])
        stats_1 = self.flatstats_to_array(ast.FlatStats(x1).calculate())
        npt.assert_allclose(stats_1, expected_stats, rtol=1e-6, equal_nan=True)

    def test_nonflat_array(self):
        x1 = None
        with pytest.raises(ValueError):
            ast.FlatStats(x1)

        x2 = np.array([[5, 9], [3, 4]])
        with pytest.raises(ValueError):
            ast.FlatStats(x2)

    def flatstats_to_array(self, fs):
        n = fs.n
        mean = fs.mean
        median = fs.median
        minimum = fs.min
        maximum = fs.max
        std = fs.std
        cv = fs.cv
        skewness = fs.skewness
        kurtosis = fs.kurtosis
        entropy = fs.entropy

        return [n, mean, median, minimum, maximum, std, cv, skewness, kurtosis,
                entropy]


class TestArrayStats:

    # Good arrays
    array_4D = np.nan*np.ones((3, 3, 2, 2))
    array_4D[:, :, 0, 0] = np.array([[4, 8, -1.2], [-4.8, 4, 2], [-3, 0, 5.0]])
    array_4D[:, :, 0, 1] = np.array([[3, 6, -3.7], [-3.1, 2, 0], [-4, 4, 4.3]])
    array_4D[:, :, 1, 0] = np.array([[2, 4, -2.2], [-1.7, 3, 3], [-2, 3, 3.2]])
    array_4D[:, :, 1, 1] = np.array([[2, 2, -3.8], [-0.9, 4, 5], [-0, 1, 5.0]])

    array_3D = np.nan*np.ones((3, 3, 2))
    array_3D[:, :, 0] = np.array([[3, 6, -3.7], [-3.1, 2, 0], [-4, 4, 4.3]])
    array_3D[:, :, 1] = np.array([[2, 4, -2.2], [-1.7, 3, 3], [-2, 3, 3.2]])

    array_2D = np.array([[3, 6, -3.7], [-3.1, 2, 0], [-4, 4, 4.3]])

    # Bad arrays
    array_5D = np.ones((2, 2, 2, 2, 2))
    array_1D = np.array([1])

    def test_4D_without_roi(self):
        # Calculate stats using ArrayStats
        stats = ast.ArrayStats(self.array_4D).calculate()

        # Calculate "gold standard" (gs) stats
        gs_4D = calc_stats(self.array_4D.flatten())
        gs_3D_0 = calc_stats(self.array_4D[:, :, :, 0].flatten())
        gs_3D_1 = calc_stats(self.array_4D[:, :, :, 1].flatten())
        gs_2D_0 = calc_stats(self.array_4D[:, :, 0, 0].flatten())
        gs_2D_1 = calc_stats(self.array_4D[:, :, 0, 1].flatten())
        gs_2D_2 = calc_stats(self.array_4D[:, :, 1, 0].flatten())
        gs_2D_3 = calc_stats(self.array_4D[:, :, 1, 1].flatten())

        gs_list = [gs_4D, gs_3D_0, gs_3D_1, gs_2D_0, gs_2D_1, gs_2D_2, gs_2D_3]

        # Perform tests without ROI
        self.group_tests_4D(stats, gs_list)

    def test_4D_with_roi(self):
        # "Random" ROI
        roi_4D = np.ones((3, 3, 2, 2), dtype=bool)
        roi_4D[:, :, 0, 0] = np.array([[0, 0, 0], [1, 1, 1], [0, 0, 0]])
        roi_4D[:, :, 0, 1] = np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0]])
        roi_4D[:, :, 1, 0] = np.array([[0, 1, 1], [0, 0, 0], [0, 0, 0]])
        roi_4D[:, :, 1, 1] = np.array([[1, 1, 0], [1, 0, 0], [0, 0, 1]])

        # Calculate stats using ArrayStats
        stats_roi = ast.ArrayStats(self.array_4D, roi=roi_4D).calculate()

        # Calculate "gold standard" (gs) stats
        gs_4D_roi = calc_stats(self.array_4D[roi_4D])
        gs_3D_roi_0 = calc_stats(self.array_4D[:, :, :, 0][roi_4D[:, :, :, 0]])
        gs_3D_roi_1 = calc_stats(self.array_4D[:, :, :, 1][roi_4D[:, :, :, 1]])
        gs_2D_roi_0 = calc_stats(self.array_4D[:, :, 0, 0][roi_4D[:, :, 0, 0]])
        gs_2D_roi_1 = calc_stats(self.array_4D[:, :, 0, 1][roi_4D[:, :, 0, 1]])
        gs_2D_roi_2 = calc_stats(self.array_4D[:, :, 1, 0][roi_4D[:, :, 1, 0]])
        gs_2D_roi_3 = calc_stats(self.array_4D[:, :, 1, 1][roi_4D[:, :, 1, 1]])

        gs_list_roi = [gs_4D_roi, gs_3D_roi_0, gs_3D_roi_1,
                       gs_2D_roi_0, gs_2D_roi_1, gs_2D_roi_2, gs_2D_roi_3]

        # Perform tests with ROI
        self.group_tests_4D(stats_roi, gs_list_roi)

    def group_tests_4D(self, stats, gs_list):
        """
        Perform tests that compare outputs of ArrayStats.calculate() and the
        calc_stats() function above.

        Notes
        -----
        This is used by tests with and without ROI so it was abstracted in its
        own function
        """

        # Unpack list
        [gs_4D, gs_3D_0, gs_3D_1, gs_2D_0, gs_2D_1, gs_2D_2, gs_2D_3] = gs_list

        # Group gs stats to compare with output from ArrayStats
        gs_n_4D = gs_4D[0]
        gs_mean_4D = gs_4D[1]
        gs_median_4D = gs_4D[2]
        gs_minimum_4D = gs_4D[3]
        gs_maximum_4D = gs_4D[4]
        gs_std_4D = gs_4D[5]
        gs_cv_4D = gs_4D[6]
        gs_skewness_4D = gs_4D[7]
        gs_kurtosis_4D = gs_4D[8]
        gs_entropy_4D = gs_4D[9]

        gs_n_3D = np.array([gs_3D_0[0], gs_3D_1[0]])
        gs_mean_3D = np.array([gs_3D_0[1], gs_3D_1[1]])
        gs_median_3D = np.array([gs_3D_0[2], gs_3D_1[2]])
        gs_minimum_3D = np.array([gs_3D_0[3], gs_3D_1[3]])
        gs_maximum_3D = np.array([gs_3D_0[4], gs_3D_1[4]])
        gs_std_3D = np.array([gs_3D_0[5], gs_3D_1[5]])
        gs_cv_3D = np.array([gs_3D_0[6], gs_3D_1[6]])
        gs_skewness_3D = np.array([gs_3D_0[7], gs_3D_1[7]])
        gs_kurtosis_3D = np.array([gs_3D_0[8], gs_3D_1[8]])
        gs_entropy_3D = np.array([gs_3D_0[9], gs_3D_1[9]])

        gs_n_2D = np.array([[gs_2D_0[0], gs_2D_1[0]],
                            [gs_2D_2[0], gs_2D_3[0]]])
        gs_mean_2D = np.array([[gs_2D_0[1], gs_2D_1[1]],
                               [gs_2D_2[1], gs_2D_3[1]]])
        gs_median_2D = np.array([[gs_2D_0[2], gs_2D_1[2]],
                                 [gs_2D_2[2], gs_2D_3[2]]])
        gs_minimum_2D = np.array([[gs_2D_0[3], gs_2D_1[3]],
                                  [gs_2D_2[3], gs_2D_3[3]]])
        gs_maximum_2D = np.array([[gs_2D_0[4], gs_2D_1[4]],
                                  [gs_2D_2[4], gs_2D_3[4]]])
        gs_std_2D = np.array([[gs_2D_0[5], gs_2D_1[5]],
                              [gs_2D_2[5], gs_2D_3[5]]])
        gs_cv_2D = np.array([[gs_2D_0[6], gs_2D_1[6]],
                             [gs_2D_2[6], gs_2D_3[6]]])
        gs_skewness_2D = np.array([[gs_2D_0[7], gs_2D_1[7]],
                                   [gs_2D_2[7], gs_2D_3[7]]])
        gs_kurtosis_2D = np.array([[gs_2D_0[8], gs_2D_1[8]],
                                   [gs_2D_2[8], gs_2D_3[8]]])
        gs_entropy_2D = np.array([[gs_2D_0[9], gs_2D_1[9]],
                                  [gs_2D_2[9], gs_2D_3[9]]])

        # Tests
        npt.assert_allclose(stats["n"]["4D"], gs_n_4D, rtol=1e-6)
        npt.assert_allclose(stats["mean"]["4D"], gs_mean_4D, rtol=1e-6)
        npt.assert_allclose(stats["median"]["4D"], gs_median_4D, rtol=1e-6)
        npt.assert_allclose(stats["min"]["4D"], gs_minimum_4D, rtol=1e-6)
        npt.assert_allclose(stats["max"]["4D"], gs_maximum_4D, rtol=1e-6)
        npt.assert_allclose(stats["std"]["4D"], gs_std_4D, rtol=1e-6)
        npt.assert_allclose(stats["cv"]["4D"], gs_cv_4D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"]["4D"], gs_skewness_4D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"]["4D"], gs_kurtosis_4D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"]["4D"], gs_entropy_4D, rtol=1e-6)

        npt.assert_allclose(stats["n"]["3D"], gs_n_3D, rtol=1e-6)
        npt.assert_allclose(stats["mean"]["3D"], gs_mean_3D, rtol=1e-6)
        npt.assert_allclose(stats["median"]["3D"], gs_median_3D, rtol=1e-6)
        npt.assert_allclose(stats["min"]["3D"], gs_minimum_3D, rtol=1e-6)
        npt.assert_allclose(stats["max"]["3D"], gs_maximum_3D, rtol=1e-6)
        npt.assert_allclose(stats["std"]["3D"], gs_std_3D, rtol=1e-6)
        npt.assert_allclose(stats["cv"]["3D"], gs_cv_3D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"]["3D"], gs_skewness_3D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"]["3D"], gs_kurtosis_3D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"]["3D"], gs_entropy_3D, rtol=1e-6)

        npt.assert_allclose(stats["n"]["2D"], gs_n_2D, rtol=1e-6)
        npt.assert_allclose(stats["mean"]["2D"], gs_mean_2D, rtol=1e-6)
        npt.assert_allclose(stats["median"]["2D"], gs_median_2D, rtol=1e-6)
        npt.assert_allclose(stats["min"]["2D"], gs_minimum_2D, rtol=1e-6)
        npt.assert_allclose(stats["max"]["2D"], gs_maximum_2D, rtol=1e-6)
        npt.assert_allclose(stats["std"]["2D"], gs_std_2D, rtol=1e-6)
        npt.assert_allclose(stats["cv"]["2D"], gs_cv_2D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"]["2D"], gs_skewness_2D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"]["2D"], gs_kurtosis_2D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"]["2D"], gs_entropy_2D, rtol=1e-6)

    def test_3D_without_roi(self):
        # Calculate stats using ArrayStats
        stats = ast.ArrayStats(self.array_3D).calculate()

        # Calculate "gold standard" (gs) stats
        gs_3D = calc_stats(self.array_3D.flatten())
        gs_2D_0 = calc_stats(self.array_3D[:, :, 0].flatten())
        gs_2D_1 = calc_stats(self.array_3D[:, :, 1].flatten())

        # Group gs stats to compare with output from ArrayStats
        gs_n_3D = gs_3D[0]
        gs_mean_3D = gs_3D[1]
        gs_median_3D = gs_3D[2]
        gs_minimum_3D = gs_3D[3]
        gs_maximum_3D = gs_3D[4]
        gs_std_3D = gs_3D[5]
        gs_cv_3D = gs_3D[6]
        gs_skewness_3D = gs_3D[7]
        gs_kurtosis_3D = gs_3D[8]
        gs_entropy_3D = gs_3D[9]

        gs_n_2D = np.array([gs_2D_0[0], gs_2D_1[0]])
        gs_mean_2D = np.array([gs_2D_0[1], gs_2D_1[1]])
        gs_median_2D = np.array([gs_2D_0[2], gs_2D_1[2]])
        gs_minimum_2D = np.array([gs_2D_0[3], gs_2D_1[3]])
        gs_maximum_2D = np.array([gs_2D_0[4], gs_2D_1[4]])
        gs_std_2D = np.array([gs_2D_0[5], gs_2D_1[5]])
        gs_cv_2D = np.array([gs_2D_0[6], gs_2D_1[6]])
        gs_skewness_2D = np.array([gs_2D_0[7], gs_2D_1[7]])
        gs_kurtosis_2D = np.array([gs_2D_0[8], gs_2D_1[8]])
        gs_entropy_2D = np.array([gs_2D_0[9], gs_2D_1[9]])

        npt.assert_allclose(stats["n"]["3D"], gs_n_3D, rtol=1e-6)
        npt.assert_allclose(stats["mean"]["3D"], gs_mean_3D, rtol=1e-6)
        npt.assert_allclose(stats["median"]["3D"], gs_median_3D, rtol=1e-6)
        npt.assert_allclose(stats["min"]["3D"], gs_minimum_3D, rtol=1e-6)
        npt.assert_allclose(stats["max"]["3D"], gs_maximum_3D, rtol=1e-6)
        npt.assert_allclose(stats["std"]["3D"], gs_std_3D, rtol=1e-6)
        npt.assert_allclose(stats["cv"]["3D"], gs_cv_3D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"]["3D"], gs_skewness_3D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"]["3D"], gs_kurtosis_3D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"]["3D"], gs_entropy_3D, rtol=1e-6)

        npt.assert_allclose(stats["n"]["2D"], gs_n_2D, rtol=1e-6)
        npt.assert_allclose(stats["mean"]["2D"], gs_mean_2D, rtol=1e-6)
        npt.assert_allclose(stats["median"]["2D"], gs_median_2D, rtol=1e-6)
        npt.assert_allclose(stats["min"]["2D"], gs_minimum_2D, rtol=1e-6)
        npt.assert_allclose(stats["max"]["2D"], gs_maximum_2D, rtol=1e-6)
        npt.assert_allclose(stats["std"]["2D"], gs_std_2D, rtol=1e-6)
        npt.assert_allclose(stats["cv"]["2D"], gs_cv_2D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"]["2D"], gs_skewness_2D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"]["2D"], gs_kurtosis_2D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"]["2D"], gs_entropy_2D, rtol=1e-6)

    def test_2D_without_roi(self):
        # Calculate stats using ArrayStats
        stats = ast.ArrayStats(self.array_2D).calculate()

        # Calculate "gold standard" (gs) stats
        gs_2D = calc_stats(self.array_2D.flatten())

        # Group gs stats to compare with output from ArrayStats
        gs_n_2D = gs_2D[0]
        gs_mean_2D = gs_2D[1]
        gs_median_2D = gs_2D[2]
        gs_minimum_2D = gs_2D[3]
        gs_maximum_2D = gs_2D[4]
        gs_std_2D = gs_2D[5]
        gs_cv_2D = gs_2D[6]
        gs_skewness_2D = gs_2D[7]
        gs_kurtosis_2D = gs_2D[8]
        gs_entropy_2D = gs_2D[9]

        npt.assert_allclose(stats["n"], gs_n_2D, rtol=1e-6)
        npt.assert_allclose(stats["mean"], gs_mean_2D, rtol=1e-6)
        npt.assert_allclose(stats["median"], gs_median_2D, rtol=1e-6)
        npt.assert_allclose(stats["min"], gs_minimum_2D, rtol=1e-6)
        npt.assert_allclose(stats["max"], gs_maximum_2D, rtol=1e-6)
        npt.assert_allclose(stats["std"], gs_std_2D, rtol=1e-6)
        npt.assert_allclose(stats["cv"], gs_cv_2D, rtol=1e-6)
        npt.assert_allclose(stats["skewness"], gs_skewness_2D, rtol=1e-6)
        npt.assert_allclose(stats["kurtosis"], gs_kurtosis_2D, rtol=1e-6)
        npt.assert_allclose(stats["entropy"], gs_entropy_2D, rtol=1e-6)

    def test_bad_array_dimensions(self):
        with pytest.raises(ValueError):
            ast.ArrayStats(self.array_1D)

        with pytest.raises(ValueError):
            ast.ArrayStats(self.array_5D)

    def test_bad_roi_dtype(self):
        array = np.ones((3, 3, 2))
        roi = np.ones((3, 3, 2))
        with pytest.raises(TypeError):
            ast.ArrayStats(array, roi)

    def test_roi_shape_mismatch(self):
        array = np.ones((3, 3, 2))
        roi = np.ones((3, 3, 3), dtype=bool)
        with pytest.raises(ValueError):
            ast.ArrayStats(array, roi)



"""
This module contains multiple auxiliary functions that might be used by
multiple algorithms

"""
import numpy as np
from scipy.ndimage import zoom


def convert_to_pi_range(pixel_array):
    """
    Rescale the image values to the interval [-pi, pi].

    Parameters
    ----------
    pixel_array : np.ndarray

    Returns
    -------
    radians_array : np.ndarray
        An array containing with the same shape as pixel_array
        scaled to the range [-pi, pi].
    """
    if (np.amax(pixel_array) > np.pi) or (np.amin(pixel_array) < -np.pi):
        pi_array = np.pi * np.ones(np.shape(pixel_array))
        min_array = np.amin(pixel_array) * np.ones(np.shape(pixel_array))
        max_array = np.amax(pixel_array) * np.ones(np.shape(pixel_array))
        radians_array = (2.0 * pi_array * (pixel_array - min_array) /
                         (max_array - min_array)) - pi_array
    else:
        # It means it's already on the interval [-pi, pi]
        radians_array = pixel_array
    return radians_array


def resize_array(pixel_array, factor=1, target_size=None):
    """
    Resizes the given pixel_array, using target_size as the reference
    of the resizing operation (if None, then there's no resizing).
    This method applies a resize_factor to the first 2 axes of the input array.
    The remaining axes are left unchanged.
    Example 1: (10, 10, 2) => (5, 5, 2) with resize_factor = 0.5
    Example 2: (10, 10, 10, 2) => (20, 20, 10, 2) with resize_factor = 2

    Parameters
    ----------
    pixel_array : np.ndarray

    factor : boolean
        Optional input argument. This is the resize factor defined by the user
        and it is applied in the scipy.ndimage.zoom

    target_size : boolean
        Optional input argument. By default, this script does not apply the
        scipy wrap_around in the input image.

    Returns
    -------
    resized_array : np.ndarray where the size of the first 2 dimensions
        is np.shape(pixel_array) * factor. The remaining dimensions (or axes)
        will have a the same size as in pixel_array.
    """
    if target_size is not None:
        factor = target_size / np.shape(pixel_array)[0]

    resize_factor = np.ones(len(np.shape(pixel_array)))
    resize_factor[0] = factor
    resize_factor[1] = factor
    resized_array = zoom(pixel_array, resize_factor)

    return resized_array


def mask_slices(shape, slices, mask=None):
    """
    Get mask to limit processing to specific slices.

    This function allows to quickly get a mask of Trues of the right shape to
    limit processing to specific slices. If `mask` is provided it outputs a new
    mask corresponding to the input mask but only on the specified slices.

    Parameters
    ----------
    shape : tuple
        shape of mask to be created
    slices : int or list (of ints)
        slice indices where mask is to be True
    mask : np.ndarray (of booleans)
        original mask, if provided this function will return a mask of Falses
        in all elements except at the locations of the True elements of this
        `mask` in the slices given by `slices`

    Returns
    -------
    np.ndarray (of booleans)
    """
    # Input type checks
    if not isinstance(shape, tuple):
        raise ValueError("`shape` must be a tuple")

    if not isinstance(slices, (int, list)):
        raise ValueError("`slices` must be an integer or a list of integers")

    # Check elements of `slices` are ints
    if (isinstance(slices, list) and
       not all(isinstance(x, int) for x in slices)):
        raise ValueError("Every element of `slices` must be an integer")

    # Check elements of `slices` within range defined by `shape`
    if isinstance(slices, int):
        s_min = slices
        s_max = slices
    elif isinstance(slices, list):
        s_min = min(slices)
        s_max = max(slices)

    if not(s_min >= 0 and s_max+1 <= shape[2]):
        msg = f"The elements of `slices` must be > 0 and <= {shape[2]-1}"
        raise ValueError(msg)

    # Ensure shape and the dimensions of mask match
    if mask is not None:
        assert isinstance(mask, np.ndarray), "`mask` must be a numpy"
        assert mask.dtype == "bool", "The elements of `mask` must be bool"
        assert (shape == mask.shape), "The shape of `mask` must match `shape`"

    # Make mask of Trues at the specified slices
    template = np.full(shape, False)
    template[:, :, slices] = True

    # If `mask` provided, apply it to the template
    if mask is None:
        final_mask = template
    else:
        final_mask = np.logical_and(mask, template)

    return final_mask



import os
import shutil
from ukat.mapping.diffusion import make_gradient_scheme, ADC, DTI
from ukat.data import fetch
from ukat.utils import arraystats
import numpy.testing as npt
import numpy as np
import pytest


class TestMakeGradientScheme:

    def test_one_bzero_true_with_bzero(self):
        bvals = [0, 5, 10]
        bvecs = [[1, 0, 1],
                 [-1, 1, 0]]
        output = make_gradient_scheme(bvals, bvecs, one_bzero=True)
        expected = (" 0.70710678          0.0   0.70710678      0\n"
                    " 0.70710678          0.0   0.70710678      5\n"
                    " 0.70710678          0.0   0.70710678     10\n"
                    "-0.70710678   0.70710678          0.0      5\n"
                    "-0.70710678   0.70710678          0.0     10")
        assert output == expected

    def test_one_bzero_true_without_bzero(self):
        bvals = [5, 10]
        bvecs = [[1, 0, 1],
                 [-1, 1, 0]]
        output = make_gradient_scheme(bvals, bvecs, one_bzero=True)
        expected = (" 0.70710678          0.0   0.70710678      0\n"
                    " 0.70710678          0.0   0.70710678      5\n"
                    " 0.70710678          0.0   0.70710678     10\n"
                    "-0.70710678   0.70710678          0.0      5\n"
                    "-0.70710678   0.70710678          0.0     10")
        assert output == expected

    def test_one_bzero_true_with_bzero_dont_normalize(self):
        bvals = [0, 5, 10]
        bvecs = [[1, 0, 1],
                 [-1, 1, 0]]
        output = make_gradient_scheme(bvals, bvecs, normalize=False,
                                      one_bzero=True)
        expected = ("          1            0            1      0\n"
                    "          1            0            1      5\n"
                    "          1            0            1     10\n"
                    "         -1            1            0      5\n"
                    "         -1            1            0     10")
        assert output == expected

    def test_one_bzero_false_with_bzero(self):
        bvals = [0, 5, 10]
        bvecs = [[1, 0, 1],
                 [-1, 1, 0]]
        output = make_gradient_scheme(bvals, bvecs, one_bzero=False)
        expected = (" 0.70710678          0.0   0.70710678      0\n"
                    " 0.70710678          0.0   0.70710678      5\n"
                    " 0.70710678          0.0   0.70710678     10\n"
                    "-0.70710678   0.70710678          0.0      0\n"
                    "-0.70710678   0.70710678          0.0      5\n"
                    "-0.70710678   0.70710678          0.0     10")
        assert output == expected

    def test_one_bzero_false_without_bzero(self):
        bvals = [5, 10]
        bvecs = [[1, 0, 1],
                 [-1, 1, 0]]
        output = make_gradient_scheme(bvals, bvecs, one_bzero=False)
        expected = (" 0.70710678          0.0   0.70710678      5\n"
                    " 0.70710678          0.0   0.70710678     10\n"
                    "-0.70710678   0.70710678          0.0      5\n"
                    "-0.70710678   0.70710678          0.0     10")
        assert output == expected


class TestADC:
    pixel_array, affine, bvals, bvecs = fetch.dwi_philips()
    pixel_array = pixel_array[35:95, 40:90, 3:6, :]
    mask = pixel_array[..., 0] > 20000

    def test_missmatched_raw_data_and_bvals(self):

        with pytest.raises(AssertionError):
            mapper = ADC(self.pixel_array, self.affine, self.bvals[:-2],
                         self.mask)

    def test_fail_to_fit(self):
        mapper = ADC(self.pixel_array[..., ::-1], self.affine, self.bvals,
                     self.mask)
        assert np.abs(mapper.adc.mean()) < 1e-6

    def test_negative_signal(self):
        gold_standard_adc = [0.000833, 0.000998, 0.0, 0.004852]
        gold_standard_adc_err = [7.819414e-05, 1.222237e-04, 0.0, 9.935775e-04]
        negateive_pixel_array = self.pixel_array.copy()
        negateive_pixel_array[:30, :, :, :] -= 40000
        mapper = ADC(negateive_pixel_array, self.affine, self.bvals)
        adc_stats = arraystats.ArrayStats(mapper.adc).calculate()
        adc_err_stats = arraystats.ArrayStats(mapper.adc_err).calculate()
        assert np.sum(mapper.adc[:30, :, :]) == 0
        npt.assert_allclose([adc_stats['mean']['3D'], adc_stats['std']['3D'],
                             adc_stats['min']['3D'], adc_stats['max']['3D']],
                            gold_standard_adc, rtol=5e-4, atol=5e-7)
        npt.assert_allclose([adc_err_stats['mean']['3D'],
                             adc_err_stats['std']['3D'],
                             adc_err_stats['min']['3D'],
                             adc_err_stats['max']['3D']],
                            gold_standard_adc_err, rtol=5e-3, atol=1e-7)

    def test_real_data(self):
        # Gold standard statistics
        gold_standard_adc = [0.001455, 0.001058, 0.0, 0.005391]
        gold_standard_adc_err = [0.000127, 0.000142, 0.0, 0.001044]
        # Test maps
        mapper = ADC(self.pixel_array, self.affine, self.bvals, self.mask)
        adc_stats = arraystats.ArrayStats(mapper.adc).calculate()
        adc_err_stats = arraystats.ArrayStats(mapper.adc_err).calculate()
        npt.assert_allclose([adc_stats['mean']['3D'], adc_stats['std']['3D'],
                             adc_stats['min']['3D'], adc_stats['max']['3D']],
                            gold_standard_adc, rtol=5e-4, atol=5e-7)
        npt.assert_allclose([adc_err_stats['mean']['3D'],
                             adc_err_stats['std']['3D'],
                             adc_err_stats['min']['3D'],
                             adc_err_stats['max']['3D']],
                            gold_standard_adc_err, rtol=5e-3, atol=1e-7)

    def test_ukrin_b(self):
        # Gold standard statistics
        gold_standard_adc = [0.001396, 0.000996, 0.0, 0.004789]
        gold_standard_adc_err = [0.000237, 0.000314, 0.0, 0.00275]
        # Test maps
        mapper = ADC(self.pixel_array, self.affine, self.bvals, self.mask,
                     ukrin_b=True)
        assert mapper.n_bvals == 4
        npt.assert_array_equal(mapper.u_bvals, np.array([0, 100, 200, 800]))
        adc_stats = arraystats.ArrayStats(mapper.adc).calculate()
        adc_err_stats = arraystats.ArrayStats(mapper.adc_err).calculate()
        npt.assert_allclose([adc_stats['mean']['3D'], adc_stats['std']['3D'],
                             adc_stats['min']['3D'], adc_stats['max']['3D']],
                            gold_standard_adc, rtol=5e-4, atol=5e-7)
        npt.assert_allclose([adc_err_stats['mean']['3D'],
                             adc_err_stats['std']['3D'],
                             adc_err_stats['min']['3D'],
                             adc_err_stats['max']['3D']],
                            gold_standard_adc_err, rtol=1e-3, atol=5e-7)

    def test_to_nifti(self):
        mapper = ADC(self.pixel_array, self.affine, self.bvals, self.mask)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='adc_test', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 3
        assert 'adc_test_adc_map.nii.gz' in output_files
        assert 'adc_test_adc_err.nii.gz' in output_files
        assert 'adc_test_mask.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='adc_test', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only adc and adc_err are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='adc_test', maps=['adc', 'adc_err'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 2
        assert 'adc_test_adc_map.nii.gz' in output_files
        assert 'adc_test_adc_err.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='adc_test', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


class TestDTI:
    pixel_array, affine, bvals, bvecs = fetch.dwi_philips()
    pixel_array = pixel_array[35:95, 40:90, 3:6, :]
    mask = pixel_array[..., 0] > 20000

    def test_missmatched_raw_data_and_bvals(self):

        with pytest.raises(AssertionError):
            mapper = DTI(self.pixel_array, self.affine, self.bvals[:-2],
                         self.bvecs, self.mask)

    def test_missmatched_raw_data_and_bvecs(self):

        with pytest.raises(AssertionError):
            mapper = DTI(self.pixel_array, self.affine, self.bvals,
                         self.bvecs[:-1, :], self.mask)

    def test_bvecs_transpose(self):

        with pytest.warns(UserWarning):
            mapper = DTI(self.pixel_array, self.affine, self.bvals,
                         self.bvecs.T, self.mask)

    def test_real_data(self):
        # Gold standard statistics
        gold_standard_md = [0.001781, 0.001567, 0.0, 0.012655]
        gold_standard_fa = [0.353293, 0.256178, 0.0, 0.999999]
        gold_standard_color_fa = [0.170594, 0.185415, 0.0, 0.968977]

        # Test maps
        mapper = DTI(self.pixel_array, self.affine, self.bvals, self.bvecs,
                     self.mask)
        md_stats = arraystats.ArrayStats(mapper.md).calculate()
        fa_stats = arraystats.ArrayStats(mapper.fa).calculate()
        color_fa_stats = arraystats.ArrayStats(mapper.color_fa).calculate()
        npt.assert_allclose([md_stats['mean']['3D'], md_stats['std']['3D'],
                             md_stats['min']['3D'], md_stats['max']['3D']],
                            gold_standard_md, rtol=1e-6, atol=1e-4)
        npt.assert_allclose([fa_stats['mean']['3D'], fa_stats['std']['3D'],
                             fa_stats['min']['3D'], fa_stats['max']['3D']],
                            gold_standard_fa, rtol=1e-6, atol=1e-4)
        npt.assert_allclose([color_fa_stats['mean']['4D'],
                             color_fa_stats['std']['4D'],
                             color_fa_stats['min']['4D'],
                             color_fa_stats['max']['4D']],
                            gold_standard_color_fa, rtol=1e-6, atol=1e-4)

    def test_ukrin_b(self):
        # Gold standard statistics
        gold_standard_md = [0.001611, 0.001437, 0.0, 0.02951]
        gold_standard_fa = [0.446954, 0.291819, 0.0, 1.0]
        gold_standard_color_fa = [0.216434, 0.219389, 0.0, 0.99972]

        # Test maps
        mapper = DTI(self.pixel_array, self.affine, self.bvals, self.bvecs,
                     self.mask, ukrin_b=True)
        assert mapper.n_bvals == 4
        npt.assert_array_equal(mapper.u_bvals, np.array([0, 100, 200, 800]))
        md_stats = arraystats.ArrayStats(mapper.md).calculate()
        fa_stats = arraystats.ArrayStats(mapper.fa).calculate()
        color_fa_stats = arraystats.ArrayStats(mapper.color_fa).calculate()
        npt.assert_allclose([md_stats['mean']['3D'], md_stats['std']['3D'],
                             md_stats['min']['3D'], md_stats['max']['3D']],
                            gold_standard_md, rtol=1e-6, atol=1e-4)
        npt.assert_allclose([fa_stats['mean']['3D'], fa_stats['std']['3D'],
                             fa_stats['min']['3D'], fa_stats['max']['3D']],
                            gold_standard_fa, rtol=1e-6, atol=1e-4)
        npt.assert_allclose([color_fa_stats['mean']['4D'],
                             color_fa_stats['std']['4D'],
                             color_fa_stats['min']['4D'],
                             color_fa_stats['max']['4D']],
                            gold_standard_color_fa, rtol=1e-6, atol=1e-4)

    def test_to_nifti(self):
        mapper = DTI(self.pixel_array, self.affine, self.bvals, self.bvecs,
                     self.mask)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='dti_test', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 4
        assert 'dti_test_md_map.nii.gz' in output_files
        assert 'dti_test_fa_map.nii.gz' in output_files
        assert 'dti_test_color_fa_map.nii.gz' in output_files
        assert 'dti_test_mask.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='dti_test', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only md and fa are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='dti_test', maps=['md', 'fa'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 2
        assert 'dti_test_md_map.nii.gz' in output_files
        assert 'dti_test_fa_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='dti_test', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import os
import warnings
import numpy as np
import nibabel as nib
import concurrent.futures
from tqdm import tqdm
from scipy.optimize import curve_fit


class T2Star:
    """
    Attributes
    ----------
    t2star_map : np.ndarray
        The estimated T2* values in ms
    t2star_err : np.ndarray
        The certainty in the fit of `t2star_map` in ms. Only returned if
        `2p_exp` method is used, otherwise is an array of nan
    m0_map : np.ndarray
        The estimated M0 values
    m0_err : np.ndarray
        The certainty in the fit of `m0_map`. Only returned if `2p_exp`
        method is used, otherwise is an array of nan
    shape : tuple
        The shape of the T2* map
    n_te : int
        The number of TE used to calculate the map
    n_vox : int
        The number of voxels in the map i.e. the product of all dimensions
        apart from TE
    """

    def __init__(self, pixel_array, echo_list, affine, mask=None,
                 method='loglin', multithread='auto'):
        """Initialise a T2Star class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            A array containing the signal from each voxel at each echo
            time with the last dimension being time i.e. the array needed to
            generate a 3D T2* map would have dimensions [x, y, z, TE].
        echo_list : list()
            An array of the echo times used for the last dimension of the
            raw data. In milliseconds.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired T2* map rather than the raw data i.e. omit the time
            dimension.
        method : {'loglin', '2p_exp'}, optional
            Default `loglin`
            The method used to estimate T2* values. 'loglin' uses a
            weighted linear fit to the natural logarithm of the
            signal. '2p_exp' fits the signal to a two parameter
            exponential (S = S0 * exp(-t / T2*)). `loglin` is far quicker
            but produces inaccurate results for T2* below 20 ms. `2p_exp` is
            accurate below 20 ms however this comes at the expense of run time.
        multithread : bool or 'auto', optional
            Default 'auto'.
            If True, fitting will be distributed over all cores available on
            the node. If False, fitting will be carried out on a single thread.
            'auto' attempts to apply multithreading where appropriate based
            on the number of voxels being fit and the method being used.
            Generally 'loglin' is quicker running single threaded  due to
            the additional overheads of multithreading while '2p_exp' is
            quicker running multithreaded for anything but small numbers of
            voxels.
        """
        # Some sanity checks
        assert (pixel_array.shape[-1]
                == len(echo_list)), 'Number of echoes does not match the ' \
                                    'number of time frames on the last axis ' \
                                    'of pixel_array'
        assert method == 'loglin' \
            or method == '2p_exp', 'method must be loglin or 2p_exp. You ' \
                                   'entered {}'.format(method)
        assert multithread is True \
            or multithread is False \
            or multithread == 'auto', 'multithreaded must be True, False or ' \
                                      'auto. You entered {}'\
                                      .format(multithread)
        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.n_te = pixel_array.shape[-1]
        self.n_vox = np.prod(self.shape)
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
            # Don't process any nan values
        self.mask[np.isnan(np.sum(pixel_array, axis=-1))] = False
        self.echo_list = echo_list
        self.method = method
        # Auto multithreading conditions
        if multithread == 'auto':
            if self.method == '2p_exp' and self.n_vox > 20:
                multithread = True
            else:
                multithread = False
        self.multithread = multithread

        # Fit data
        self.t2star_map, self.t2star_err, self.m0_map, self.m0_err\
            = self.__fit__()

        # Warn if using loglin method to produce a map with a large
        # proportion of T2* < 20 ms i.e. where loglin isn't as accurate.
        if self.method == 'loglin':
            proportion_less_than_20 = np.sum((self.t2star_map > 0) &
                                             (self.t2star_map < 20)) \
                                      / np.prod(self.n_vox)
            warn_thresh = 0.3
            if proportion_less_than_20 > warn_thresh:
                warnings.warn('{:%} of voxels in this map have a T2* less '
                              'than 20 ms. The loglin method is not accurate '
                              'in this regime. If these voxels are of '
                              'interest, consider using the 2p_exp fitting'
                              ' method'.format(proportion_less_than_20))

    def __fit__(self):

        # Initialise maps
        t2star_map = np.zeros(self.n_vox)
        t2star_err = np.zeros(self.n_vox)
        m0_map = np.zeros(self.n_vox)
        m0_err = np.zeros(self.n_vox)
        mask = self.mask.flatten()
        signal = self.pixel_array.reshape(-1, self.n_te)
        # Get indices of voxels to process
        idx = np.argwhere(mask).squeeze()

        # Multithreaded method
        if self.multithread:
            with concurrent.futures.ProcessPoolExecutor() as pool:
                with tqdm(total=idx.size) as progress:
                    futures = []

                    for ind in idx:
                        future = pool.submit(self.__fit_signal__,
                                             signal[ind, :],
                                             self.echo_list,
                                             self.method)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                    results = []
                    for future in futures:
                        result = future.result()
                        results.append(result)
            t2star_map[idx], t2star_err[idx], m0_map[idx], m0_err[idx] \
                = [np.array(row) for row in zip(*results)]

        # Single threaded method
        else:
            with tqdm(total=idx.size) as progress:
                for ind in idx:
                    sig = signal[ind, :]
                    t2star_map[ind], t2star_err[ind], \
                        m0_map[ind], m0_err[ind] \
                        = self.__fit_signal__(sig, self.echo_list, self.method)
                    progress.update(1)

        # Reshape results to raw data shape
        t2star_map = t2star_map.reshape(self.shape)
        t2star_err = t2star_err.reshape(self.shape)
        m0_map = m0_map.reshape(self.shape)
        m0_err = m0_err.reshape(self.shape)

        return t2star_map, t2star_err, m0_map, m0_err

    @staticmethod
    def __fit_signal__(sig, te, method):
        if method == 'loglin':
            s_w = 0.0
            s_wx = 0.0
            s_wx2 = 0.0
            s_wy = 0.0
            s_wxy = 0.0
            n_te = len(sig)

            noise = sig.sum() / n_te
            sd = np.abs(np.sum(sig ** 2) / n_te - noise ** 2)
            if sd > 1e-10:
                for t in range(n_te):
                    if sig[t] > 0:
                        te_tmp = te[t]
                        if sig[t] > sd:
                            sigma = np.log(sig[t] / (sig[t] - sd))
                        else:
                            sigma = np.log(sig[t] / 0.0001)
                        logsig = np.log(sig[t])
                        weight = 1 / sigma ** 2

                        s_w += weight
                        s_wx += weight * te_tmp
                        s_wx2 += weight * te_tmp ** 2
                        s_wy += weight * logsig
                        s_wxy += weight * te_tmp * logsig

                delta = (s_w * s_wx2) - (s_wx ** 2)
                if delta > 1e-5:
                    a = (1 / delta) * (s_wx2 * s_wy - s_wx * s_wxy)
                    b = (1 / delta) * (s_w * s_wxy - s_wx * s_wy)
                    t2star = np.real(-1 / b)
                    m0 = np.real(np.exp(a))
                    if t2star < 0 or t2star > 700 or np.isnan(t2star):
                        t2star = 0
                        m0 = 0
                else:
                    t2star = 0
                    m0 = 0
            else:
                t2star = 0
                m0 = 0
            t2star_err = np.nan
            m0_err = np.nan

        elif method == '2p_exp':
            # Initialise parameters
            bounds = ([0, 0], [700, 1000000])
            initial_guess = [20, 10000]

            # Fit data to equation
            try:
                popt, pcov = curve_fit(two_param_eq, te, sig,
                                       p0=initial_guess, bounds=bounds)
            except RuntimeError:
                popt = np.zeros(2)
                pcov = np.zeros((2, 2))

            # Extract fits and errors from result variables
            if popt[0] < bounds[1][0] - 1:
                t2star = popt[0]
                m0 = popt[1]
                err = np.sqrt(np.diag(pcov))
                t2star_err = err[0]
                m0_err = err[1]
            else:
                t2star, m0, t2star_err, m0_err = 0, 0, 0, 0

        return t2star, t2star_err, m0, m0_err

    def r2star_map(self):
        """
        Generates the R2* map from the T2* map output by initialising this
        class.

        Parameters
        ----------
        See class attributes in __init__

        Returns
        -------
        r2star_map : np.ndarray
            An array containing the R2* map generated
            by the function with R2* measured in ms.
        """
        return np.nan_to_num(np.reciprocal(self.t2star_map),
                             posinf=0, neginf=0)

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports some of the T2Star class attributes to NIFTI.
                
        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["t2star", "t2star_err", "m0",
            "m0_err", "r2star", "mask"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['t2star', 'm0', 'r2star', 'mask']
            if self.method == '2p_exp':
                maps += ['t2star_err', 'm0_err']
        if isinstance(maps, list):
            for result in maps:
                if result == 't2star' or result == 't2star_map':
                    t2star_nifti = nib.Nifti1Image(self.t2star_map,
                                                   affine=self.affine)
                    nib.save(t2star_nifti, base_path + '_t2star_map.nii.gz')
                elif result == 't2star_err' or result == 't2star_err_map':
                    t2star_err_nifti = nib.Nifti1Image(self.t2star_err,
                                                       affine=self.affine)
                    nib.save(t2star_err_nifti, base_path +
                             '_t2star_err.nii.gz')
                    if self.method == 'loglin':
                        warnings.warn('Saving t2star_error however, '
                                      'the loglin method does not produce '
                                      'confidence intervals. As such the '
                                      'resulting nifti will be all zeros.')
                elif result == 'm0' or result == 'm0_map':
                    m0_nifti = nib.Nifti1Image(self.m0_map, affine=self.affine)
                    nib.save(m0_nifti, base_path + '_m0_map.nii.gz')
                elif result == 'm0_err' or result == 'm0_err_map':
                    m0_err_nifti = nib.Nifti1Image(self.m0_err,
                                                   affine=self.affine)
                    nib.save(m0_err_nifti, base_path + '_m0_err.nii.gz')
                    if self.method == 'loglin':
                        warnings.warn('Saving m0_error however, the loglin '
                                      'method does not produce confidence '
                                      'intervals. As such the resulting nifti'
                                      ' will be all zeros.')
                elif result == 'r2star' or result == 'r2star_map':
                    r2star_nifti = nib.Nifti1Image(T2Star.r2star_map(self),
                                                   affine=self.affine)
                    nib.save(r2star_nifti, base_path + '_r2star_map.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["t2star", "t2star_err", "m0", "m0_err", '
                             '"r2star", "mask"]".')

        return


def two_param_eq(t, t2star, m0):
    """
        Calculate the expected signal from the equation
        signal = M0 * exp(-t / T2*)

        Parameters
        ----------
        t: list
            The times the signal will be calculated at
        t2star: float
            The T2* of the signal
        m0: float
            The M0 of the signal

        Returns
        -------
        signal: np.ndarray
        """
    return np.sqrt(np.square(m0 * np.exp(-t / t2star)))



import os
import shutil
import numpy as np
import numpy.testing as npt
import pytest
from ukat.data import fetch
from ukat.mapping.mtr import MTR
from ukat.utils import arraystats


class TestMTR:
    # Create arrays for testing
    array_one = np.arange(0, 300, 3).reshape((10, 10, 1))
    array_one[-1] = 0
    array_two = np.arange(0, 200, 2).reshape((10, 10, 1))
    array_two[-1] = 0
    correct_array = np.concatenate((array_one, array_two), axis=2)
    swapped_mts_array = np.concatenate((array_two, array_one), axis=2)
    three_mts_array = np.concatenate((array_one, array_two, array_two), axis=2)
    affine = np.eye(4)

    # Gold standard: [mean, std, min, max] of MTR when input = `correct_array`
    # MTR for comparison is = `(array_two - array_one) / array_one`
    gold_standard_mtr = [0.2966666666, 0.1042965856, 0.0, 0.3333333333]

    def test_mtr_calculation(self):
        mapper = MTR(self.correct_array, self.affine)
        mtr_map_calculated = mapper.mtr_map
        mtrmaps_stats = arraystats.ArrayStats(mtr_map_calculated).calculate()
        npt.assert_allclose([mtrmaps_stats["mean"], mtrmaps_stats["std"],
                            mtrmaps_stats["min"], mtrmaps_stats["max"]],
                            self.gold_standard_mtr, rtol=1e-7, atol=1e-9)
        # The purpose of the next assert is to test the infinite situation
        assert (mtr_map_calculated[-1] ==
                np.zeros(np.shape(mtr_map_calculated[-1]))).all()
        assert (mapper.mt_off == np.squeeze(self.array_one)).all()
        assert (mapper.mt_on == np.squeeze(self.array_two)).all()

    def test_inputs(self):
        # Check that it fails when input pixel_array has incorrect shape
        with pytest.raises(AssertionError):
            MTR(self.array_one, self.affine)
        with pytest.raises(AssertionError):
            MTR(self.three_mts_array, self.affine)
        # Check if there's a warning if sum(mt_off) > sum(mt_on)
        with pytest.warns(UserWarning):
            MTR(self.swapped_mts_array, self.affine)

    def test_mask(self):
        # Create a mask
        mask = np.ones(self.correct_array.shape[:-1], dtype=bool)
        mask[:5] = False

        all_pixels = MTR(self.correct_array, self.affine)
        masked_pixels = MTR(self.correct_array, self.affine, mask=mask)

        assert (all_pixels.mtr_map != masked_pixels.mtr_map).any()
        assert (all_pixels.mt_on != masked_pixels.mt_on).any()
        assert (all_pixels.mt_off != masked_pixels.mt_off).any()
        assert (arraystats.ArrayStats(all_pixels.mtr_map).calculate() !=
                arraystats.ArrayStats(masked_pixels.mtr_map).calculate())

    def test_to_nifti(self):
        # Create a MTR map instance and test different export to NIFTI options.
        mapper = MTR(self.correct_array, self.affine)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='mtrtest', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 4
        assert 'mtrtest_mtr_map.nii.gz' in output_files
        assert 'mtrtest_mt_on.nii.gz' in output_files
        assert 'mtrtest_mt_off.nii.gz' in output_files
        assert 'mtrtest_mask.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='mtrtest', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only `mtr` and `mt_on` are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='mtrtest', maps=['mtr', 'mt_on'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 2
        assert 'mtrtest_mtr_map.nii.gz' in output_files
        assert 'mtrtest_mt_on.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper = MTR(self.correct_array, self.affine)
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='mtrtest', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')

    def test_real_data(self):
        # Get test data
        images, affine = fetch.mtr_philips()

        # Gold standard statistics
        gold_standard_mtr_real = [0.1845690591, 0.6237606679, -73.0, 1.0]
        # The minimum should be 0, but the MT_ON and MT_OFF of real data
        # isn't perfectly aligned, which will result in outliers.

        # MTR Map
        mapper = MTR(images, affine)

        # Stats comparison
        mtrmap_stats = arraystats.ArrayStats(mapper.mtr_map).calculate()
        npt.assert_allclose([mtrmap_stats["mean"], mtrmap_stats["std"],
                            mtrmap_stats["min"], mtrmap_stats["max"]],
                            gold_standard_mtr_real, rtol=0.01, atol=0)


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



"""
The velocity and flow calculations in this Phase Contrast class
are based on the scientific paper:

"Phase-contrast magnetic resonance imaging to assess renal perfusion:
a systematic review and statement paper"

Giulia Villa, Steffen Ringgaard, Ingo Hermann, Rebecca Noble, Paolo Brambilla,
Dinah S. Khatir, Frank G. Zöllner, Susan T. Francis, Nicholas M. Selby,
Andrea Remuzzi, Anna Caroli

Magnetic Resonance Materials in Physics, Biology and Medicine (2020) 33:3-21
https://doi.org/10.1007/s10334-019-00772-0
"""

import os
import pandas as pd
import numpy as np
import nibabel as nib
import matplotlib.pyplot as plt
from tabulate import tabulate
from ukat.utils.tools import convert_to_pi_range


class PhaseContrast:
    """
    Generates velocity and flow measurements of an MRI Phase Contrast Sequence.

    Attributes
    ----------
    velocity_array : np.ndarray
        The input velocity array masked.
    shape : tuple
        The shape of the velocity array.
    pixel_spacing : tuple
        The pixel spacing of the acquisition estimated from the affine.
    mask : np.ndarray
        A boolean mask of the voxels to fit.
    num_pixels : np.ndarray
        List containing the number of True values in the mask per phase.
    area : np.ndarray
        List containing the area (cm2) of the mask per phase.
    min_velocity : np.ndarray
        List containing the minimum velocity values (cm/s) per phase.
    mean_velocity : np.ndarray
        List containing the average velocity values (cm/s) per phase.
    max_velocity : np.ndarray
        List containing the maximum velocity values (cm/s) per phase.
    std_velocity : np.ndarray
        List containing the std dev of the velocity values (cm/s) per phase.
    rbf : np.ndarray
        List containing the Renal Blood Flow values (ml/min) per phase.
    mean_velocity_global : float
        Average velocity (cm/s) across the different phases.
    mean_rbf : float
        Average Renal Blood Flow (ml/min) across the different phases.
    resistive_index : float
        A prognostic marker in renal vascular diseases which range is [0, 1].
    """

    def __init__(self, velocity_array, affine, mask):
        """Initialise a PhaseContrast class instance.

        Parameters
        ----------
        velocity_array : np.ndarray
            A 3D array containing the velocity images of the phase contrast
            sequence, i.e. the dimensions of the array are [x, y, p], where p
            corresponds to the phase (or trigger delay).
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray
            A boolean mask of the voxels to fit. Should be the shape of
            the raw data.
        """
        self.shape = velocity_array.shape
        self.affine = affine
        self.pixel_spacing = (np.linalg.norm(self.affine[:3, 1]),
                              np.linalg.norm(self.affine[:3, 0]))
        self.mask = np.where(mask, mask, np.nan)
        self.velocity_array = np.abs(velocity_array * self.mask)
        self.num_pixels = []
        self.area = []
        self.min_velocity = []
        self.mean_velocity = []
        self.max_velocity = []
        self.std_velocity = []
        self.rbf = []
        self.mean_velocity_global = 0
        self.mean_rbf = 0
        self.resistive_index = 0
        if len(self.shape) == 3:
            # Extract number pixels, area, velocity stats (min, mean, max, std)
            # and renal blood flow (RBF)
            self.num_pixels = np.count_nonzero(~np.isnan(self.velocity_array),
                                               axis=(0, 1))
            # area = (num_pixels * mm * mm * 0.01) = cm2
            self.area = (self.num_pixels * self.pixel_spacing[0] *
                         self.pixel_spacing[1] * 0.01)
            self.min_velocity = np.nanmin(self.velocity_array, axis=(0, 1))
            self.mean_velocity = np.nanmean(self.velocity_array, axis=(0, 1))
            self.max_velocity = np.nanmax(self.velocity_array, axis=(0, 1))
            self.std_velocity = np.nanstd(self.velocity_array, axis=(0, 1))
            # q = (60 * cm2 * cm/s) = ml/min
            self.rbf = 60 * np.array(self.area * self.mean_velocity)
            # Mean velocity global and mean flow
            self.mean_velocity_global = np.mean(self.mean_velocity)
            self.mean_rbf = np.mean(self.rbf)
            # Restrictive Index
            mean_velocity_systole = np.max(self.mean_velocity)
            mean_velocity_diastole = np.min(self.mean_velocity)
            self.resistive_index = ((mean_velocity_systole -
                                     mean_velocity_diastole) /
                                    mean_velocity_systole)
            # Convert any nan values to 0
            self.velocity_array = np.nan_to_num(self.velocity_array)
            self.mask = np.nan_to_num(self.mask)
        else:
            raise ValueError('The input velocity_array should be 3D.')

    def get_stats_table(self):
        """
        Stores most of PhaseContrast class attributes into a pandas DataFrame.

        Returns
        ----------
        table : pandas.DataFrame
            Returns a table with the results/stats of each output per phase.
        """
        stats = {"RBF (ml/min)": self.rbf,
                 "Area (cm2)": self.area,
                 "Nr Pixels": self.num_pixels,
                 "Mean Vel (cm/s)": self.mean_velocity,
                 "Min Vel (cm/s)": self.min_velocity,
                 "Max Vel (cm/s)": self.max_velocity,
                 "StdDev Vel (cm/s)": self.std_velocity}
        table = pd.DataFrame(data=stats)
        return table

    def print_stats(self):
        """
        Prints the table with the stats for each output per phase.
        """
        stats_table = self.get_stats_table()
        print(tabulate(stats_table, headers='keys', tablefmt='github',
              floatfmt='.3f'))

    def to_csv(self, path):
        """
        Saves the stats_table into a csv file.

        Parameters
        ----------
        path : str
            Path to the desired csv file.
        """
        stats_table = self.get_stats_table()
        stats_table.to_csv(path)

    def plot(self, stat='default', file_name=None):
        """
        This method plots the output PhaseContrast stats per phase.

        Parameters
        ----------
        stat : str, optional
            Name of the output stat variable. This method plots mean velocity
            and the RBF by default.
        file_name : str, optional
            Path to the image file (*.jpg, *.png, etc.)
            in which the plot will be saved.
        """
        if stat == 'default':
            fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10))
            ax1.plot(self.mean_velocity, 'ro-')
            ax1.set_ylabel('Velocity (cm/sec)')
            ax1.set_xlabel('Phase')
            ax1.set_title('Average Velocity')
            ax2.plot(self.rbf, 'b-')
            ax2.set_ylabel('RBF (ml/min)')
            ax2.set_xlabel('Phase')
            ax2.set_title('Renal Artery Blood Flow')
        else:
            if stat == 'min_velocity':
                stat_variable = self.min_velocity
                y_label = 'Velocity (cm/sec)'
                title = 'Minimum Velocity'
            elif stat == 'mean_velocity':
                stat_variable = self.mean_velocity
                y_label = 'Velocity (cm/sec)'
                title = 'Average Velocity'
            elif stat == 'max_velocity':
                stat_variable = self.max_velocity
                y_label = 'Velocity (cm/sec)'
                title = 'Maximum Velocity'
            elif stat == 'std_velocity':
                stat_variable = self.std_velocity
                y_label = 'Velocity (cm/sec)'
                title = 'Standard Deviation of the Velocity'
            elif stat == 'rbf':
                stat_variable = self.rbf
                y_label = 'RBF (ml/min)'
                title = 'Renal Artery Blood Flow'
            elif stat == 'num_pixels':
                stat_variable = self.num_pixels
                y_label = '# Pixels'
                title = 'Number of Pixels in the Region Of Interest (ROI)'
            elif stat == 'area':
                stat_variable = self.area
                y_label = 'Area ($cm^2$)'
                title = 'Area of the Region Of Interest (ROI)'
            else:
                raise ValueError('The stat provided is not valid.')
            fig, ax = plt.subplots(figsize=(10, 10))
            ax.plot(stat_variable, 'ro-')
            ax.set_ylabel(y_label)
            ax.set_xlabel('Phase')
            ax.set_title(title)
        # The following saves the plot(s) to file_name, if given.
        if file_name is not None:
            fig.savefig(file_name)

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports the velocity array and the renal artery mask to NIfTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIfTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIfTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIfTI. This should either the string "all"
            or a list of maps from ["phase_array", "mask", "velocity_array"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ["velocity_array", "mask"]
        if isinstance(maps, list):
            for result in maps:
                if result == 'velocity_array' or result == 'velocity array':
                    velocity_nifti = nib.Nifti1Image(self.velocity_array,
                                                     affine=self.affine)
                    nib.save(velocity_nifti, base_path +
                             '_velocity_array.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIfTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["velocity_array", "mask"]".')


def convert_to_velocity(pixel_array, velocity_encoding,
                        velocity_encode_scale=None):
    """
    Calculate the velocity array from the given input image and
    velocity encoding. If a velocity encode scale is given then it is
    used to convert the pixel_value to radians.

    Parameters
    ----------
    pixel_array: np.ndarray
        A 3D array containing the phase images of the phase contrast sequence.
    velocity_encoding : float
        The value of the velocity encoding in cm/s.
    velocity_encode_scale : float, optional
        If given, this value is used to scale from image intensity to radians.

    Returns
    -------
    velocity_array: np.ndarray
        A 3D array containing the velocity images of the phase contrast
        sequence, i.e. the dimensions of the array are [x, y, p], where p
        corresponds to the phase (or trigger delay).
    """
    if velocity_encode_scale is not None:
        pi_range_array = pixel_array / velocity_encode_scale
    else:
        pi_range_array = convert_to_pi_range(pixel_array)
    velocity_array = pi_range_array * velocity_encoding
    return velocity_array



import csv
import os
import shutil
import numpy.testing as npt
import pytest

from ukat.data import fetch
from ukat.segmentation.whole_kidney import Segmentation
from ukat.utils import arraystats


class TestSegmentation:
    image, affine = fetch.t2w_volume_philips()
    segmentation = Segmentation(image, affine)

    def test_get_mask(self):
        expected = [0.022764, 0.14915, 0.0, 1.0]
        mask = self.segmentation.get_mask()
        mask_stats = arraystats.ArrayStats(mask).calculate()
        npt.assert_allclose([mask_stats["mean"]["3D"], mask_stats["std"]["3D"],
                             mask_stats["min"]["3D"], mask_stats["max"]["3D"]],
                            expected, rtol=1e-6, atol=1e-4)

    def test_get_kidneys(self):
        expected = [0.034317, 0.237162, 0.0, 2.0]
        mask = self.segmentation.get_kidneys()
        mask_stats = arraystats.ArrayStats(mask).calculate()
        npt.assert_allclose([mask_stats["mean"]["3D"], mask_stats["std"]["3D"],
                             mask_stats["min"]["3D"], mask_stats["max"]["3D"]],
                            expected, rtol=1e-6, atol=1e-4)

    def test_get_left_kidney(self):
        expected = [0.011211, 0.105285, 0.0, 1.0]
        mask = self.segmentation.get_left_kidney()
        mask_stats = arraystats.ArrayStats(mask).calculate()
        npt.assert_allclose([mask_stats["mean"]["3D"], mask_stats["std"]["3D"],
                             mask_stats["min"]["3D"], mask_stats["max"]["3D"]],
                            expected, rtol=1e-6, atol=1e-4)

    def test_get_right_kidney(self):
        expected = [0.011553, 0.106863, 0.0, 1.0]
        mask = self.segmentation.get_right_kidney()
        mask_stats = arraystats.ArrayStats(mask).calculate()
        npt.assert_allclose([mask_stats["mean"]["3D"], mask_stats["std"]["3D"],
                             mask_stats["min"]["3D"], mask_stats["max"]["3D"]],
                            expected, rtol=1e-6, atol=1e-4)

    def test_get_volumes(self):
        expected = {'total': 240.00054654884337,
                    'left': 118.19352480602264,
                    'right': 121.80702174282074}
        volumes = self.segmentation.get_volumes()
        assert volumes == expected

    def test_get_tkv(self):
        expected = 240.00054654884337
        assert self.segmentation.get_tkv() == expected

    def test_get_lkv(self):
        expected = 118.19352480602264
        assert self.segmentation.get_lkv() == expected

    def test_get_rkv(self):
        expected = 121.80702174282074
        assert self.segmentation.get_rkv() == expected

    def test_save_volumes_csv(self):
        expected = [['total', 'left', 'right'],
                    ['240.00054654884337',
                     '118.19352480602264',
                     '121.80702174282074']]

        os.makedirs('test_output', exist_ok=True)
        self.segmentation.save_volumes_csv('test_output/volumes.csv')
        output_files = os.listdir('test_output')

        assert 'volumes.csv' in output_files

        with open('test_output/volumes.csv', 'r') as csv_file:
            reader = csv.reader(csv_file)
            list_from_csv = [row for row in reader]
        assert list_from_csv == expected

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))
        shutil.rmtree('test_output')

    def test_to_nifti(self):
        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        self.segmentation.to_nifti(output_directory='test_output',
                                   base_file_name='test_subject', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 4
        assert 'test_subject_mask.nii.gz' in output_files
        assert 'test_subject_left_kidney.nii.gz' in output_files
        assert 'test_subject_right_kidney.nii.gz' in output_files
        assert 'test_subject_individual_kidneys.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        with pytest.raises(ValueError):
            self.segmentation.to_nifti(output_directory='test_output',
                                       base_file_name='test_subject',
                                       maps='not_an_option')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import pytest
import numpy as np
import numpy.testing as npt
import ukat.utils.tools as tools
from ukat.utils import arraystats


class TestConvertToPiRange:

    # Gold Standard = [mean, std, minimum, maximum]
    # Input: {np.arange(12).reshape((2, 2, 3)) - 6 * np.ones((2, 2, 3))}
    gold_standard = [-7.401486830834377e-17, 1.9718077939258474,
                     -3.141592653589793, 3.1415926535897922]

    # Create arrays for testing
    array = np.arange(12).reshape((2, 2, 3)) - 6 * np.ones((2, 2, 3))
    array_positive = np.arange(12).reshape((2, 2, 3))
    array_negative = -np.arange(12).reshape((2, 2, 3))
    array_pi_range = array / 2

    def test_pi_range_result(self):
        pi_range_calculated = tools.convert_to_pi_range(self.array)
        stats = arraystats.ArrayStats(pi_range_calculated).calculate()
        npt.assert_allclose([stats["mean"]["3D"], stats["std"]["3D"],
                                    stats["min"]["3D"], stats["max"]["3D"]],
                                    self.gold_standard, rtol=1e-6, atol=1e-4)

    def test_if_ranges(self):
        # Test for values > 3.2
        result_over = tools.convert_to_pi_range(self.array_positive)
        assert np.amax(result_over) < np.amax(self.array_positive)
        # Test for values < -3.2
        result_under = tools.convert_to_pi_range(self.array_negative)
        assert np.amin(result_under) > np.amin(self.array_negative)
        # Test for values > 3.2 and < -3.2
        result_under_over = tools.convert_to_pi_range(self.array)
        assert np.amax(result_under_over) < np.amax(self.array)
        assert np.amin(result_under_over) > np.amin(self.array)

    def test_else_range(self):
        result = tools.convert_to_pi_range(self.array_pi_range)
        assert (result == self.array_pi_range).all()

    def test_input_array_type_assertion(self):
        # Empty array
        with pytest.raises(ValueError):
            tools.convert_to_pi_range(np.array([]))
        # No input argument
        with pytest.raises(TypeError):
            tools.convert_to_pi_range(None)
        # String
        with pytest.raises(TypeError):
            tools.convert_to_pi_range("abcdef")


class TestResizeArray:

    # Create arrays for testing
    array_2d = np.arange(100).reshape((10, 10))
    array_3d = np.arange(500).reshape((10, 10, 5))
    array_4d = np.arange(5000).reshape((10, 10, 5, 10))

    def test_no_resize(self):
        resized_array_2d = tools.resize_array(self.array_2d)
        resized_array_3d = tools.resize_array(self.array_3d, factor=1)
        resized_array_4d = tools.resize_array(self.array_4d, target_size=10)
        assert (resized_array_2d == self.array_2d).all()
        assert (resized_array_3d == self.array_3d).all()
        assert (resized_array_4d == self.array_4d).all()

    def test_output_shapes(self):
        resized_array_1 = tools.resize_array(self.array_3d, factor=2)
        resized_array_2 = tools.resize_array(self.array_4d, target_size=20)
        assert np.shape(resized_array_1)[0] != 10
        assert np.shape(resized_array_1)[0] == 20
        assert np.shape(resized_array_1)[0] == np.shape(resized_array_1)[1]
        assert np.shape(resized_array_2)[0] != 10
        assert np.shape(resized_array_2)[0] == 20
        assert np.shape(resized_array_1)[0] == np.shape(resized_array_2)[0]
        assert np.shape(resized_array_1)[1] == np.shape(resized_array_2)[1]
        assert np.shape(resized_array_1)[2] != 10
        assert np.shape(resized_array_1)[2] == np.shape(self.array_3d)[2]
        assert np.shape(resized_array_2)[3] != 20
        assert np.shape(resized_array_2)[3] == np.shape(self.array_4d)[3]

    def test_input_array_type_assertion(self):
        # Empty array
        with pytest.raises(IndexError):
            tools.resize_array(np.array([]))
        # No input argument
        with pytest.raises(IndexError):
            tools.resize_array(None)
        # String
        with pytest.raises(IndexError):
            tools.resize_array("abcdef")


class TestMaskSlices:

    shape = (2, 2, 3)
    # Create mask where all pixels from all slices are True
    full_mask = np.full(shape, True)
    # Create mask where only the pixels from slice index 1 are True
    one_slice_mask = np.full(shape, False)
    one_slice_mask[:, :, 1] = True
    # Create mask where only the pixels from slices index 1 and 2 are True
    two_slice_mask = np.full(shape, False)
    two_slice_mask[:, :, 1] = True
    two_slice_mask[:, :, 2] = True

    def test_single_slice(self):
        # #1: shape + single slice (int)
        final_mask = tools.mask_slices(self.shape, 1)
        assert (final_mask == self.one_slice_mask).all()

    def test_multiple_slices(self):
        # #2: shape + multiple slices (list)
        final_mask = tools.mask_slices(self.shape, [1, 2])
        assert (final_mask == self.two_slice_mask).all()

    def test_masked_single_slice(self):
        # #3: shape + single slice + mask
        final_mask = tools.mask_slices(self.shape, 1, self.full_mask)
        assert (final_mask == self.one_slice_mask).all()

    def test_masked_masked_slices(self):
        # #4: shape + multiple slices + mask
        final_mask = tools.mask_slices(self.shape, [1, 2], self.full_mask)
        assert (final_mask == self.two_slice_mask).all()

    def test_shape_assertion(self):
        # Wrong type `shape`
        non_tuple_shape = [2, 2, 3]
        with pytest.raises(ValueError):
            tools.mask_slices(non_tuple_shape, 1)
        # `mask` with different dimensions than `shape`
        mismatched_mask = np.full((2, 3, 3), True)
        with pytest.raises(AssertionError):
            tools.mask_slices(self.shape, 1, mismatched_mask)

    def test_slices_type_assertion(self):
        # Wrong type `slices`
        with pytest.raises(ValueError):
            tools.mask_slices(self.shape, np.array([1, 2]))
        # Not all elements of slices are `ints`
        with pytest.raises(ValueError):
            tools.mask_slices(self.shape, [1, 2.2])
        with pytest.raises(ValueError):
            tools.mask_slices(self.shape, 2.2)

    def test_slice_ranges_assertion(self):
        # Slices out of range tests
        with pytest.raises(ValueError):
            tools.mask_slices(self.shape, [0, 3])
        with pytest.raises(ValueError):
            tools.mask_slices(self.shape, 3)

    def test_mask_type_assertion(self):
        # Wrong dtype mask
        wrong_type_mask = "not a numpy array"
        with pytest.raises(AssertionError):
            tools.mask_slices(self.shape, 1, wrong_type_mask)

    def test_mask_dtype_assertion(self):
        # Wrong dtype mask
        wrong_dtype_mask = np.full(self.shape, 2)
        with pytest.raises(AssertionError):
            tools.mask_slices(self.shape, 1, wrong_dtype_mask)



def scale_b1(data, flip_angle):
    """
    Scale the B1 map produced by GE scanners to be a percentage of the
    nominal flip angle.

    Its advisable to verify the velocity encode scale (dicom tag (0019,
    10E2)) is 299.97.

    Parameters
    ----------
    data : np.ndarray
        The B1 map to be scaled.
    flip_angle : float
        The flip angle of the sequence.

    Returns
    -------
    scaled_data : np.ndarray
        The scaled B1 map.
    """
    norm_factor = 10 / flip_angle
    b1_scaled = data * norm_factor
    return b1_scaled



import numpy as np


def write_dvs(scheme, filepath, normalization='none',
              coordinate_system='xyz', comment='N/A'):
    """Create Siemens .dvs file for diffusion imaging acquisitions

    Parameters
    ----------
    scheme : string
        Scheme can be two types of strings:
        1) File path to a text file containing a gradient scheme (see below)
        2) A gradient scheme string (see below)

        The following is an example of a dummy gradient scheme initialised in
        the `SCHEME` varible:
            >> print(SCHEME)
             0.70710678          0.0   0.70710678      0
             0.70710678          0.0   0.70710678      5
             0.70710678          0.0   0.70710678    100
             0.70710678          0.0   0.70710678    500
            -0.70710678   0.70710678          0.0      5
            -0.70710678   0.70710678          0.0    100
            -0.70710678   0.70710678          0.0    500

        Where each row corresponds to one measurement, the first 3 columns are
        the b-vector components (b-vector with norm 1) and the 4th column is
        the b-value in s/mm2. This example shows an acquisition with 3 nonzero
        b-values (5, 100, 500), each with two directions, and one b=0 in the
        start of the acquisition. Such a gradient scheme string can be
        generated using ukat.dwi.make_gradient_scheme().
    filepath: string
        The filepath (without extension) for the .dvs file which will be
        generated by this function

    Returns
    -------
    string
        filepath of generated file, including extension
    string
        content of generated file

    Notes
    -----
    The specification for Siemens .dvs files implemented here is described in
    the Siemens syngo MR E11 Operator Manual page 140. It may not work for
    other software lines (e.g. VX*).
    Note that the validity in the formatting of the `scheme` string is not
    checked so ensure you have generated it properly as shown above.

    """

    expected_normalization = ["maximum", "unity", "none"]
    if normalization not in expected_normalization:
        raise ValueError((f"`normalization` must be one of "
                          f"{expected_normalization}"))

    expected_coordinate_system = ["xyz", "prs"]
    if coordinate_system not in expected_coordinate_system:
        raise ValueError((f"`coordinate_system` must be one of "
                          f"{expected_coordinate_system}"))

    # Quick check just to keep comments short
    good_comment = isinstance(comment, str) and len(comment) < 50
    if not good_comment:
        raise ValueError("`comment` must be a string with < 50 characters")

    if ".txt" in scheme:
        # Assume is filepath
        gt = np.loadtxt(scheme)
    else:
        # Assume is string such as those generated by make_gradient_scheme()
        gt = np.vstack([np.fromstring(x, dtype=float, sep=' ')
                        for x in scheme.split('\n')])

    bvals = gt[:, 3].astype(int)
    bvecs = gt[:, :3]
    n_bvecs = bvecs.shape[0]
    unique_bvals = np.unique(bvals)
    max_bval = np.max(unique_bvals)

    # Write custom header for reference (ignored by the scanner)
    dvs = ""
    head = ("# -----------------------------------------------------\n"
            "# .dvs file for the following setup:\n"
            "# b-val (s/mm2)  # directions\n")
    dvs = f"{dvs}{head}"

    for unique_bval in unique_bvals:
        cndirs = str(np.count_nonzero(bvals == unique_bval))
        dvs = f"{dvs}# {str(unique_bval).ljust(15)}{cndirs}\n"

    dvs = f"{dvs}# -----------------------------------------------------\n"

    # .dvs directives
    dvs = f"{dvs}\n[directions={n_bvecs}]\n"
    dvs = f"{dvs}Normalization = {normalization}\n"
    dvs = f"{dvs}CoordinateSystem = {coordinate_system}\n"
    dvs = f"{dvs}Comment = {comment}\n\n"

    # Write .dvs gradient table with scaled vectors
    for i, (bval, bvec) in enumerate(zip(bvals, bvecs)):
        x, y, z = bvec*np.sqrt(bval/max_bval)
        dvs = f"{dvs}Vector[{i}] = ( {x:.6f}, {y:.6f}, {z:.6f} )\n"

    # Write b-value table for reference (ignored by the scanner)
    for i, bval in enumerate(bvals):
        dvs = f"{dvs}\n# b-value[{i}] = {bval}"

    filepath = f"{filepath}.dvs"
    fid = open(filepath, "w")
    fid.write(dvs)
    fid.close()

    return filepath, dvs



import numpy as np
import numpy.testing as npt
import os
import shutil
from ukat.data import fetch
from ukat.qa import snr
from ukat.utils import arraystats


class TestIsnr:

    def test_automatic_masking(self):
        # T2W data
        gold_standard = [0.235846, 0.424526, 0.0, 1.0]
        data, affine = fetch.t2w_volume_philips()
        isnr_obj = snr.Isnr(data)
        assert isnr_obj.noise_mask.shape == data.shape
        isnr_stats = arraystats.ArrayStats(isnr_obj.noise_mask).calculate()
        npt.assert_allclose([isnr_stats['mean']['3D'], isnr_stats['std']['3D'],
                             isnr_stats['min']['3D'], isnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)
        npt.assert_allclose(isnr_obj.isnr, 45.968827)

        # T1W data
        gold_standard = [0.251823, 0.43406, 0.0, 1.0]
        data, affine = fetch.t1w_volume_philips()
        isnr_obj = snr.Isnr(data)
        assert isnr_obj.noise_mask.shape == data.shape
        isnr_stats = arraystats.ArrayStats(isnr_obj.noise_mask).calculate()
        npt.assert_allclose([isnr_stats['mean']['3D'], isnr_stats['std']['3D'],
                             isnr_stats['min']['3D'], isnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)
        npt.assert_allclose(isnr_obj.isnr, 57.847968)

        # T2star data
        gold_standard = [0.497915, 0.499996, 0.0, 1.0]
        data, affine, te = fetch.t2star_philips()
        isnr_obj = snr.Isnr(data)
        assert isnr_obj.noise_mask.shape == data.shape
        isnr_stats = arraystats.ArrayStats(isnr_obj.noise_mask).calculate()
        npt.assert_allclose([isnr_stats['mean']['4D'], isnr_stats['std']['4D'],
                             isnr_stats['min']['4D'], isnr_stats['max']['4D']],
                            gold_standard, rtol=1e-6, atol=1e-4)
        npt.assert_allclose(isnr_obj.isnr, 17.096477)

    def test_manual_noise_mask(self):
        np.random.seed(0)
        background = np.random.randn(128, 64, 5) * 25 + 50
        signal = np.random.randn(128, 64, 5) * 150 + 1000
        data = np.concatenate((background, signal), axis=1)

        mask = np.ones(data.shape, dtype=bool)
        mask[:, 64:, :] = False

        isnr_obj = snr.Isnr(data, mask)
        assert isnr_obj.noise_mask.shape == data.shape
        npt.assert_allclose(isnr_obj.isnr, 26.307718)

    def test_different_clusters(self):
        data, affine = fetch.t2w_volume_philips()

        # Three components (default)
        gold_standard_bg = [0.235846, 0.424526, 0.0, 1.0]
        gold_standard_clusters = [0.936891, 0.636081, 0.0, 2.0]
        isnr_obj = snr.Isnr(data)
        assert isnr_obj.noise_mask.shape == data.shape
        isnr_stats_bg = arraystats.ArrayStats(isnr_obj.noise_mask).calculate()
        npt.assert_allclose([isnr_stats_bg['mean']['3D'],
                             isnr_stats_bg['std']['3D'],
                             isnr_stats_bg['min']['3D'],
                             isnr_stats_bg['max']['3D']],
                            gold_standard_bg, rtol=1e-6, atol=1e-4)
        isnr_stats_clusters = arraystats.ArrayStats(
            isnr_obj.clusters).calculate()
        npt.assert_allclose([isnr_stats_clusters['mean']['3D'],
                             isnr_stats_clusters['std']['3D'],
                             isnr_stats_clusters['min']['3D'],
                             isnr_stats_clusters['max']['3D']],
                            gold_standard_clusters, rtol=1e-6, atol=1e-4)
        npt.assert_allclose(isnr_obj.isnr, 45.968827)

        # Four components
        gold_standard_bg = [0.240496, 0.427385, 0.0, 1.0]
        gold_standard_clusters = [1.164344, 0.994117, 0.0, 3.0]
        isnr_obj = snr.Isnr(data, n_clusters=4)
        assert isnr_obj.noise_mask.shape == data.shape
        isnr_stats_bg = arraystats.ArrayStats(isnr_obj.noise_mask).calculate()
        npt.assert_allclose([isnr_stats_bg['mean']['3D'],
                             isnr_stats_bg['std']['3D'],
                             isnr_stats_bg['min']['3D'],
                             isnr_stats_bg['max']['3D']],
                            gold_standard_bg, rtol=1e-6, atol=1e-4)
        isnr_stats_clusters = arraystats.ArrayStats(
            isnr_obj.clusters).calculate()
        npt.assert_allclose([isnr_stats_clusters['mean']['3D'],
                             isnr_stats_clusters['std']['3D'],
                             isnr_stats_clusters['min']['3D'],
                             isnr_stats_clusters['max']['3D']],
                            gold_standard_clusters, rtol=1e-6, atol=1e-4)
        npt.assert_allclose(isnr_obj.isnr, 42.08993)


class TestTsnr:
    np.random.seed(0)
    signal_high = np.random.randn(128, 64, 5, 40) * 50 + 9000
    signal_low = np.random.randn(128, 64, 5, 40) * 150 + 9000
    data = np.concatenate((signal_high, signal_low), axis=1)
    affine = np.eye(4)

    def test_tsnr_synthetic_data(self):
        gold_standard = [127.408817, 65.935981, 42.293117, 370.549169]

        tsnr_obj = snr.Tsnr(self.data, self.affine)
        assert tsnr_obj.shape == self.data.shape[:-1]
        assert tsnr_obj.n_d == self.data.shape[-1]
        tsnr_stats = arraystats.ArrayStats(tsnr_obj.tsnr_map).calculate()
        npt.assert_allclose([tsnr_stats['mean']['3D'], tsnr_stats['std']['3D'],
                             tsnr_stats['min']['3D'], tsnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)

    def test_mask(self):
        gold_standard = [95.566893, 96.903421, 0.0, 370.549169]

        # Bool mask
        mask = np.ones(self.data.shape[:-1], dtype=bool)
        mask[:, 64:, :] = False
        tsnr_obj = snr.Tsnr(self.data, self.affine, mask=mask)
        assert tsnr_obj.shape == self.data.shape[:-1]
        assert tsnr_obj.n_d == self.data.shape[-1]
        tsnr_stats = arraystats.ArrayStats(tsnr_obj.tsnr_map).calculate()
        npt.assert_allclose([tsnr_stats['mean']['3D'], tsnr_stats['std']['3D'],
                             tsnr_stats['min']['3D'], tsnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)

        # Int mask
        mask = np.ones(self.data.shape[:-1])
        mask[:, 64:, :] = 0
        tsnr_obj = snr.Tsnr(self.data, self.affine, mask=mask)
        assert tsnr_obj.shape == self.data.shape[:-1]
        assert tsnr_obj.n_d == self.data.shape[-1]
        tsnr_stats = arraystats.ArrayStats(tsnr_obj.tsnr_map).calculate()
        npt.assert_allclose([tsnr_stats['mean']['3D'], tsnr_stats['std']['3D'],
                             tsnr_stats['min']['3D'], tsnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)

    def test_real_data(self):
        # High tSNR sample data
        gold_standard = [9.436505, 8.775681, -1.163641, 116.984336]
        data, affine = fetch.tsnr_high_philips()
        tsnr_obj = snr.Tsnr(data, affine)
        assert tsnr_obj.shape == data.shape[:-1]
        assert tsnr_obj.n_d == data.shape[-1]
        tsnr_stats = arraystats.ArrayStats(tsnr_obj.tsnr_map).calculate()
        npt.assert_allclose([tsnr_stats['mean']['3D'], tsnr_stats['std']['3D'],
                             tsnr_stats['min']['3D'], tsnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)

        # Low tSNR sample data
        gold_standard = [1.805342, 1.435092, -3.471621, 26.160435]
        data, affine = fetch.tsnr_low_philips()
        tsnr_obj = snr.Tsnr(data, affine)
        assert tsnr_obj.shape == data.shape[:-1]
        assert tsnr_obj.n_d == data.shape[-1]
        tsnr_stats = arraystats.ArrayStats(tsnr_obj.tsnr_map).calculate()
        npt.assert_allclose([tsnr_stats['mean']['3D'], tsnr_stats['std']['3D'],
                             tsnr_stats['min']['3D'], tsnr_stats['max']['3D']],
                            gold_standard, rtol=1e-6, atol=1e-4)

    def test_to_nifti(self):
        tsnr_obj = snr.Tsnr(self.data, self.affine)
        os.makedirs('test_output', exist_ok=True)
        tsnr_obj.to_nifti('test_output', base_file_name='synthetic_data')
        output_files = os.listdir('test_output')
        assert len(output_files) == 1
        assert 'synthetic_data_tsnr_map.nii.gz' in output_files

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import os
import shutil
import numpy as np
import numpy.testing as npt
import pytest
from ukat.data import fetch
from ukat.mapping.t1 import T1, magnitude_correct, two_param_eq, \
    two_param_abs_eq, three_param_eq, three_param_abs_eq
from ukat.utils import arraystats
from ukat.utils.tools import convert_to_pi_range


class TestT1:
    t1 = 1000
    m0 = 5000
    eff = 1.8
    t = np.linspace(200, 1000, 9)

    # The ideal signal produced by the equation M0 * (1 - 2 * exp(-t / T1))
    # where M0 = 5000 and T1 = 1000 at 9 t between 200 and 1000 ms
    correct_signal_two_param = np.array([-3187.30753078, -2408.18220682,
                                         -1703.20046036, -1065.30659713,
                                         -488.11636094, 34.14696209,
                                         506.71035883, 934.30340259,
                                         1321.20558829])
    # The ideal signal produced by the equation M0 * (1 - eff * exp(-t /
    # T1)) where M0 = 5000, eff = 1.8 and T1 = 1000 at 9 t between 200
    # and 1000 ms
    correct_signal_three_param = np.array([-2368.5767777, -1667.36398614,
                                           -1032.88041432, -458.77593741,
                                           60.69527515, 530.73226588,
                                           956.03932295, 1340.87306233,
                                           1689.08502946])
    # The ideal signal produced by the equation M0 * (1 - 2 * exp(-t / T1))
    # where M0 = 5000 and T1 = 1000 acquired over three slices at 9 t
    # between 200 and 1000 ms + a temporal slice spacing of 10 ms
    correct_signal_two_param_tss = np.array([[[-3187.30753078, -2408.18220682,
                                               -1703.20046036, -1065.30659713,
                                               -488.11636094, 34.14696209,
                                               506.71035883, 934.30340259,
                                               1321.20558829],
                                              [-3105.8424597, -2334.46956224,
                                               -1636.50250136, -1004.95578812,
                                               -433.50869074, 83.55802539,
                                               551.41933777, 974.75775966,
                                               1357.81020428],
                                              [-3025.18797962, -2261.49037074,
                                               -1570.46819815, -945.2054797,
                                               -379.44437595, 132.4774404,
                                               595.68345494, 1014.80958915,
                                               1394.05059827]],
                                             [[-3187.30753078, -2408.18220682,
                                               -1703.20046036, -1065.30659713,
                                               -488.11636094, 34.14696209,
                                               506.71035883, 934.30340259,
                                               1321.20558829],
                                              [-3105.8424597, -2334.46956224,
                                               -1636.50250136, -1004.95578812,
                                               -433.50869074, 83.55802539,
                                               551.41933777, 974.75775966,
                                               1357.81020428],
                                              [-3025.18797962, -2261.49037074,
                                               -1570.46819815, -945.2054797,
                                               -379.44437595, 132.4774404,
                                               595.68345494, 1014.80958915,
                                               1394.05059827]]
                                             ])
    # Signal with all 9 elements equal to -5000 between 200 and 1000 ms
    signal_fail_fit = -5000 * np.ones(9)
    affine = np.eye(4)

    def test_two_param_eq(self):
        # Without abs
        signal = two_param_eq(self.t, self.t1, self.m0)
        npt.assert_allclose(signal, self.correct_signal_two_param,
                            rtol=1e-6, atol=1e-8)
        # With abs
        signal = two_param_abs_eq(self.t, self.t1, self.m0)
        npt.assert_allclose(signal, np.abs(self.correct_signal_two_param),
                            rtol=1e-6, atol=1e-8)

    def test_three_param_eq(self):
        # Without abs
        signal = three_param_eq(self.t, self.t1, self.m0, self.eff)
        npt.assert_allclose(signal, self.correct_signal_three_param,
                            rtol=1e-6, atol=1e-8)
        # With abs
        signal = three_param_abs_eq(self.t, self.t1, self.m0, self.eff)
        npt.assert_allclose(signal, np.abs(self.correct_signal_three_param),
                            rtol=1e-6, atol=1e-8)

    def test_two_param_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal_two_param, (10, 10, 3, 1))

        # Multithread
        mapper = T1(signal_array, self.t, self.affine, multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

        # Single Threaded
        mapper = T1(signal_array, self.t, self.affine, multithread=False)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

    def test_three_param_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal_three_param, (10, 10, 3, 1))

        # Multithread
        mapper = T1(signal_array, self.t, self.affine, parameters=3,
                    multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.eff_map.mean() - self.eff < 0.00005
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

        # Single Threaded
        mapper = T1(signal_array, self.t, self.affine, parameters=3,
                    multithread=False)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.eff_map.mean() - self.eff < 0.00005
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

    def test_tss(self):

        mapper = T1(self.correct_signal_two_param_tss, self.t, self.affine,
                    tss=10)
        assert mapper.shape == self.correct_signal_two_param_tss.shape[:-1]
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

    def test_tss_axis(self):
        signal_array = np.swapaxes(self.correct_signal_two_param_tss, 0, 1)
        mapper = T1(signal_array, self.t, self.affine, tss=10, tss_axis=0)
        assert mapper.t1_map.mean() - self.t1 < 0.00001
        assert mapper.m0_map.mean() - self.m0 < 0.00001
        assert mapper.r1_map().mean() - 1 / self.t1 < 0.00001

    def test_failed_fit(self):
        # Make the signal, where the fitting is expected to fail, into 4D array
        signal_array = np.tile(self.signal_fail_fit, (10, 10, 3, 1))

        # Fail to fit using the 2 parameter equation
        mapper_two_param = T1(signal_array, self.t, self.affine,
                              parameters=2, multithread=True)
        assert mapper_two_param.shape == signal_array.shape[:-1]
        # Voxels that fail to fit are set to zero
        assert mapper_two_param.t1_map.mean() == 0.0
        assert mapper_two_param.t1_err.mean() == 0.0
        assert mapper_two_param.m0_map.mean() == 0.0
        assert mapper_two_param.m0_err.mean() == 0.0

        # Fail to fit using the 3 parameter equation
        mapper_three_param = T1(signal_array, self.t, self.affine,
                                parameters=3, multithread=True)
        assert mapper_three_param.shape == signal_array.shape[:-1]
        # Voxels that fail to fit are set to zero
        assert mapper_three_param.t1_map.mean() == 0.0
        assert mapper_three_param.t1_err.mean() == 0.0
        assert mapper_three_param.m0_map.mean() == 0.0
        assert mapper_three_param.m0_err.mean() == 0.0

    def test_mask(self):
        signal_array = np.tile(self.correct_signal_two_param, (10, 10, 3, 1))

        # Bool mask
        mask = np.ones(signal_array.shape[:-1], dtype=bool)
        mask[:5, ...] = False
        mapper = T1(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map[5:, ...].mean() - self.t1 < 0.00001
        assert mapper.t1_map[:5, ...].mean() < 0.00001

        # Int mask
        mask = np.ones(signal_array.shape[:-1])
        mask[:5, ...] = 0
        mapper = T1(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        assert mapper.t1_map[5:, ...].mean() - self.t1 < 0.00001
        assert mapper.t1_map[:5, ...].mean() < 0.00001

    def test_mismatched_raw_data_and_inversion_lengths(self):

        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 4)),
                        inversion_list=np.linspace(0, 2000, 5),
                        affine=self.affine)

        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5)),
                        inversion_list=np.linspace(0, 2000, 4),
                        affine=self.affine)

    def test_parameters(self):

        # One parameter fit
        with pytest.raises(ValueError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5)),
                        inversion_list=np.linspace(0, 2000, 5),
                        affine=self.affine, parameters=1)

        # Four parameter fit
        with pytest.raises(ValueError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5)),
                        inversion_list=np.linspace(0, 2000, 5),
                        affine=self.affine, parameters=4)

    def test_tss_valid_axis(self):

        # 4D -1 index
        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5, 10)),
                        inversion_list=np.linspace(0, 2000, 10),
                        affine=self.affine, tss=1, tss_axis=-1)

        # 4D 3 index
        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5, 10)),
                        inversion_list=np.linspace(0, 2000, 10),
                        affine=self.affine, tss=1, tss_axis=3)

        # 4D 4 index
        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 5, 10)),
                        inversion_list=np.linspace(0, 2000, 10),
                        affine=self.affine, tss=1, tss_axis=4)

        # 3D -1 index
        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 10)),
                        inversion_list=np.linspace(0, 2000, 10),
                        affine=self.affine, tss=1, tss_axis=-1)

        # 3D 2 index
        with pytest.raises(AssertionError):
            mapper = T1(pixel_array=np.zeros((5, 5, 10)),
                        inversion_list=np.linspace(0, 2000, 10),
                        affine=self.affine, tss=1, tss_axis=2)

    def test_molli_2p_warning(self):
        with pytest.warns(UserWarning):
            mapper = T1(pixel_array=np.zeros((5, 5, 5)),
                        inversion_list=np.linspace(0, 2000, 5),
                        affine=self.affine, parameters=2, molli=True)

    def test_real_data(self):
        # Get test data
        magnitude, phase, affine, ti, tss = fetch.t1_philips(2)
        image_molli, affine_molli, ti_molli = fetch.t1_molli_philips()

        # Convert times to ms
        ti = np.array(ti) * 1000
        tss *= 1000
        ti_molli *= 1000

        # Crop to reduce runtime
        magnitude = magnitude[37:55, 65:85, :2, :]
        image_molli = image_molli[37:55, 65:85, :2, :]

        # Gold standard statistics
        gold_standard_2p = [1041.581031, 430.129308, 241.512336, 2603.911794]
        gold_standard_3p = [1416.989523, 722.097507, 0.0, 4909.693108]
        gold_standard_3p_single = [1379.242715, 714.21752, 0.0, 4308.23814]
        gold_standard_molli = [782.923767, 495.751163, 0.0, 4452.606435]

        # Two parameter method
        mapper = T1(magnitude, ti, affine, parameters=2, tss=tss)
        t1_stats = arraystats.ArrayStats(mapper.t1_map).calculate()
        npt.assert_allclose([t1_stats['mean']['3D'], t1_stats['std']['3D'],
                             t1_stats['min']['3D'], t1_stats['max']['3D']],
                            gold_standard_2p, rtol=1e-6, atol=1e-4)

        # Three parameter method
        mapper = T1(magnitude, ti, affine, parameters=3, tss=tss)
        t1_stats = arraystats.ArrayStats(mapper.t1_map).calculate()
        npt.assert_allclose([t1_stats['mean']['3D'], t1_stats['std']['3D'],
                             t1_stats['min']['3D'], t1_stats['max']['3D']],
                            gold_standard_3p, rtol=1e-4, atol=5e-2)

        # Three parameter method for first slice only
        mapper = T1(magnitude[:, :, 0, :], ti, affine, parameters=3,
                    tss=0, tss_axis=None)
        t1_stats = arraystats.ArrayStats(mapper.t1_map).calculate()
        npt.assert_allclose([t1_stats['mean'], t1_stats['std'],
                             t1_stats['min'], t1_stats['max']],
                            gold_standard_3p_single, rtol=1e-6, atol=5e-2)

        # MOLLI corrections/data
        mapper = T1(image_molli, ti_molli, affine_molli, parameters=3,
                    molli=True)
        t1_stats = arraystats.ArrayStats(mapper.t1_map).calculate()
        npt.assert_allclose([t1_stats['mean']['3D'], t1_stats['std']['3D'],
                             t1_stats['min']['3D'], t1_stats['max']['3D']],
                            gold_standard_molli, rtol=1e-6, atol=5e-3)

    def test_to_nifti(self):
        # Create a T1 map instance and test different export to NIFTI scenarios
        signal_array = np.tile(self.correct_signal_three_param, (10, 10, 3, 1))
        mapper = T1(signal_array, self.t, self.affine, parameters=3)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t1test', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 8
        assert 't1test_eff_err.nii.gz' in output_files
        assert 't1test_eff_map.nii.gz' in output_files
        assert 't1test_m0_err.nii.gz' in output_files
        assert 't1test_m0_map.nii.gz' in output_files
        assert 't1test_mask.nii.gz' in output_files
        assert 't1test_r1_map.nii.gz' in output_files
        assert 't1test_t1_err.nii.gz' in output_files
        assert 't1test_t1_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t1test', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only t1, r1 and efficiency are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t1test', maps=['t1', 'r1', 'eff'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 3
        assert 't1test_t1_map.nii.gz' in output_files
        assert 't1test_r1_map.nii.gz' in output_files
        assert 't1test_eff_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper = T1(signal_array, self.t, self.affine)
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='t1test', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


class TestMagnitudeCorrect:

    real = np.array([-43611, -46086, -19840, -14032, 8654])
    imag = np.array([51432, 30621, 5189, 4677, -6265])

    # Numpy complex data
    comp = real + imag * (0 + 1j)
    # "Simple data" uses the last dimension to hold real and imaginary
    # components respectively
    simple = np.array([real, imag]).T

    correct_array = np.array([-67432.70678981, -55331.41094351,
                              -20507.34797579, -14790.92130329,
                              10683.72318061])

    def test_complex_conversion_shape(self):

        # Has no effect on already complex data
        corrected = magnitude_correct(self.comp)
        assert corrected.shape == (5,)

        # Converts the last dimension to complex data
        corrected = magnitude_correct(self.simple)
        assert corrected.shape == (5,)

        # Raise error if not complex data but last dimension doesn't have
        # length two i.e. isn't real and imag
        with pytest.raises(ValueError):
            corrected = magnitude_correct(self.simple[:, 0])

    def test_input_dimensions(self):

        # Tile existing data to increase dimensions
        # Comp tested up to 4D i.e. [x, y, z, TI]
        comp_2d = np.tile(self.comp, (4, 1))
        comp_3d = np.tile(self.comp, (4, 4, 1))
        comp_4d = np.tile(self.comp, (4, 4, 4, 1))
        # Simple tested up to 5D i.e. [x, y, z, TI, re/im]
        simp_3d = np.tile(self.simple, (4, 1, 1))
        simp_4d = np.tile(self.simple, (4, 4, 1, 1))
        simp_5d = np.tile(self.simple, (4, 4, 4, 1, 1))

        corrected = magnitude_correct(comp_2d)
        assert corrected.shape == (4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 1)),
                            rtol=1e-9, atol=1e-9)

        corrected = magnitude_correct(comp_3d)
        assert corrected.shape == (4, 4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 4, 1)),
                            rtol=1e-9, atol=1e-9)

        corrected = magnitude_correct(comp_4d)
        assert corrected.shape == (4, 4, 4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 4, 4, 1)),
                            rtol=1e-9, atol=1e-9)

        corrected = magnitude_correct(simp_3d)
        assert corrected.shape == (4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 1)),
                            rtol=1e-9, atol=1e-9)

        corrected = magnitude_correct(simp_4d)
        assert corrected.shape == (4, 4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 4, 1)),
                            rtol=1e-9, atol=1e-9)

        corrected = magnitude_correct(simp_5d)
        assert corrected.shape == (4, 4, 4, 5)
        npt.assert_allclose(corrected,
                            np.tile(self.correct_array, (4, 4, 4, 1)),
                            rtol=1e-9, atol=1e-9)

    def test_real_data(self):
        # Get test data
        magnitude, phase, affine, ti, tss = fetch.t1_philips(2)
        phase = convert_to_pi_range(phase)

        gold_standard = [59.35023, 191.800416, -2381.208749, 4347.05731]
        # Convert magnitude and phase into complex data
        complex_data = magnitude * (np.cos(phase) + 1j * np.sin(phase))

        magnitude_corrected = np.nan_to_num(magnitude_correct(complex_data))

        complex_stats = arraystats.ArrayStats(magnitude_corrected).calculate()
        npt.assert_allclose([complex_stats['mean']['4D'],
                             complex_stats['std']['4D'],
                             complex_stats['min']['4D'],
                             complex_stats['max']['4D']],
                            gold_standard, rtol=1e-6, atol=1e-4)


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import numpy as np
import pytest

from ukat.data import fetch


class TestFetch:

    def test_ge_b0(self):
        # Test if the fetch function works
        magnitude, phase, affine, echo_times = fetch.b0_ge()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

    def test_philips_b0(self):

        # Test if the fetch function works
        magnitude, phase, affine, echo_times = fetch.b0_philips()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

    def test_siemens_b0(self):
        # Test if the fetch function works
        magnitude, phase, affine, echo_times = fetch.b0_siemens(1)

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

        # Test if the fetch function works
        magnitude, phase, affine, echo_times = fetch.b0_siemens(2)

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

        # If no dataset_id is given
        with pytest.raises(TypeError):
            magnitude, phase, affine, echo_times = fetch.b0_siemens()

        # If an incorrect dataset_id is given
        with pytest.raises(ValueError):
            magnitude, phase, affine, echo_times = fetch.b0_siemens(3)

    def test_ge_dwi(self):
        # Test if the fetch function works
        magnitude, affine, bvals, bvecs = fetch.dwi_ge()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(bvals, np.ndarray)
        assert isinstance(bvecs, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(bvals)) == 1
        assert len(np.shape(bvecs)) == 2
        assert (np.shape(bvecs)[0] == 3 or np.shape(bvecs)[1] == 3)

    def test_philips_dwi(self):

        # Test if the fetch function works
        magnitude, affine, bvals, bvecs = fetch.dwi_philips()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(bvals, np.ndarray)
        assert isinstance(bvecs, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(bvals)) == 1
        assert len(np.shape(bvecs)) == 2
        assert (np.shape(bvecs)[0] == 3 or np.shape(bvecs)[1] == 3)

    def test_siemens_dwi(self):
        # Test if the fetch function works
        magnitude, affine, bvals, bvecs = fetch.dwi_siemens()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(bvals, np.ndarray)
        assert isinstance(bvecs, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(bvals)) == 1
        assert len(np.shape(bvecs)) == 2
        assert (np.shape(bvecs)[0] == 3 or np.shape(bvecs)[1] == 3)

    def test_philips_mtr(self):
        # Test if the fetch function works
        images, affine = fetch.mtr_philips()

        # Check the format of the outputs
        assert isinstance(images, np.ndarray)
        assert np.unique(np.isnan(images)) != [True]
        assert isinstance(affine, np.ndarray)
        assert np.shape(images)[-1] == 2
        assert len(np.shape(images)) == 3
        assert np.shape(affine) == (4, 4)

    def test_philips_pc_left(self):
        # Test if the fetch function works
        magnitude, phase, mask, affine, \
            velocity_encoding = fetch.phase_contrast_left_philips()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(mask, np.ndarray)
        assert len(np.unique(mask)) == 2
        assert isinstance(affine, np.ndarray)
        assert isinstance(velocity_encoding, int)
        assert len(np.shape(magnitude)) == 3
        assert len(np.shape(phase)) == 3
        assert len(np.shape(mask)) == 3
        assert np.shape(affine) == (4, 4)

    def test_philips_pc_right(self):
        # Test if the fetch function works
        magnitude, phase, mask, affine, \
            velocity_encoding = fetch.phase_contrast_right_philips()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(mask, np.ndarray)
        assert len(np.unique(mask)) == 2
        assert isinstance(affine, np.ndarray)
        assert isinstance(velocity_encoding, int)
        assert len(np.shape(magnitude)) == 3
        assert len(np.shape(phase)) == 3
        assert len(np.shape(mask)) == 3
        assert np.shape(affine) == (4, 4)

    def test_philips_t1(self):
        # Test if the fetch function works
        magnitude, phase, affine, inversion_times, tss = fetch.t1_philips(1)

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(inversion_times, np.ndarray)
        assert isinstance(tss, int)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(inversion_times)) == 1

        # Test if the fetch function works
        magnitude, phase, affine, inversion_times, tss = fetch.t1_philips(2)

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(phase, np.ndarray)
        assert np.unique(np.isnan(phase)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(inversion_times, np.ndarray)
        assert isinstance(tss, float)
        assert len(np.shape(magnitude)) == 4
        assert len(np.shape(phase)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(inversion_times)) == 1

        # If no dataset_id is given
        with pytest.raises(TypeError):
            magnitude, phase, affine, inversion_times, _ = \
                fetch.t1_philips()

        # If an incorrect dataset_id is given
        with pytest.raises(ValueError):
            magnitude, phase, affine, inversion_times, _ = \
                fetch.t1_philips(3)

    def test_philips_t1_molli(self):
        # Test if the fetch function works
        image, affine, inversion_times = fetch.t1_molli_philips()

        # Check the format of the outputs
        assert isinstance(image, np.ndarray)
        assert np.unique(np.isnan(image)) != [True]
        assert len(np.shape(image)) == 4
        assert isinstance(affine, np.ndarray)
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(inversion_times)) == 1

    def test_philips_t1w(self):
        # Test if the fetch function works
        image, affine = fetch.t1w_volume_philips()

        # Check the format of the outputs
        assert isinstance(image, np.ndarray)
        assert np.unique(np.isnan(image)) != [True]
        assert isinstance(affine, np.ndarray)
        assert len(np.shape(image)) == 3
        assert np.shape(affine) == (4, 4)

    def test_philips_t2(self):
        # Test if the fetch function works
        magnitude, affine, echo_times = fetch.t2_philips(1)

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

        # If an incorrect dataset_id is given
        with pytest.raises(ValueError):
            magnitude, affine, echo_times = fetch.t2_philips(3)

    def test_ge_t2star(self):
        # Test if the fetch function works
        magnitude, affine, echo_times = fetch.t2star_ge()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

    def test_philips_t2star(self):
        # Test if the fetch function works
        magnitude, affine, echo_times = fetch.t2star_philips()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

    def test_siemens_t2star(self):
        # Test if the fetch function works
        magnitude, affine, echo_times = fetch.t2star_siemens()

        # Check the format of the outputs
        assert isinstance(magnitude, np.ndarray)
        assert np.unique(np.isnan(magnitude)) != [True]
        assert isinstance(affine, np.ndarray)
        assert isinstance(echo_times, np.ndarray)
        assert len(np.shape(magnitude)) == 4
        assert np.shape(affine) == (4, 4)
        assert len(np.shape(echo_times)) == 1

    def test_philips_t2w(self):
        # Test if the fetch function works
        image, affine = fetch.t2w_volume_philips()

        # Check the format of the outputs
        assert isinstance(image, np.ndarray)
        assert np.unique(np.isnan(image)) != [True]
        assert isinstance(affine, np.ndarray)
        assert len(np.shape(image)) == 3
        assert np.shape(affine) == (4, 4)

    def test_philips_tsnr(self):
        # Test if the fetch function works with high tSNR data
        data, affine = fetch.tsnr_high_philips()

        # Check the format of the outputs
        assert isinstance(data, np.ndarray)
        assert np.unique(np.isnan(data)) != [True]
        assert isinstance(affine, np.ndarray)
        assert len(np.shape(data)) == 4
        assert np.shape(affine) == (4, 4)

        # Test if the fetch function works with low tSNR data
        data, affine = fetch.tsnr_low_philips()

        # Check the format of the outputs
        assert isinstance(data, np.ndarray)
        assert np.unique(np.isnan(data)) != [True]
        assert isinstance(affine, np.ndarray)
        assert len(np.shape(data)) == 4
        assert np.shape(affine) == (4, 4)



import os
import shutil
import numpy as np
import numpy.testing as npt
import pytest
from ukat.data import fetch
from ukat.mapping.b0 import B0
from ukat.utils import arraystats


class TestB0:
    # Create arrays for testing
    correct_array = np.arange(200).reshape((10, 10, 2))
    # `correct_array` is wrapped using the algorithm in
    # https://scikit-image.org/docs/dev/auto_examples/filters/plot_phase_unwrap.html
    correct_array = np.angle(np.exp(1j * correct_array))
    one_echo_array = np.arange(100).reshape((10, 10, 1))
    multiple_echoes_array = (np.concatenate((correct_array,
                             np.arange(300).reshape((10, 10, 3))), axis=2))
    affine = np.eye(4)
    correct_echo_list = [4, 7]
    one_echo_list = [4]
    multiple_echo_list = [1, 2, 3, 4, 5]

    # Gold standard: [mean, std, min, max] of B0 when input = `correct_array`
    gold_standard = [13.051648, 108.320512, -280.281686, 53.051648]

    def test_b0_calculation_without_unwrapping(self):
        b0_map_calculated = B0(self.correct_array,
                               self.correct_echo_list, self.affine,
                               unwrap=False).b0_map
        b0maps_stats = arraystats.ArrayStats(b0_map_calculated).calculate()
        npt.assert_allclose([b0maps_stats["mean"], b0maps_stats["std"],
                            b0maps_stats["min"], b0maps_stats["max"]],
                            self.gold_standard, rtol=1e-7, atol=1e-9)

    def test_inputs(self):
        # Check that it fails when input pixel_array has incorrect shape
        with pytest.raises(ValueError):
            B0(self.one_echo_array, self.correct_echo_list, self.affine)
        with pytest.raises(ValueError):
            B0(self.multiple_echoes_array, self.correct_echo_list, self.affine)

        # Check that it fails when input echo_list has incorrect shape
        with pytest.raises(ValueError):
            B0(self.correct_array, self.one_echo_list, self.affine)
        with pytest.raises(ValueError):
            B0(self.correct_array, self.multiple_echo_list, self.affine)

        # And when both input pixel_array and echo_list have incorrect shapes
        with pytest.raises(ValueError):
            B0(self.one_echo_array, self.one_echo_list, self.affine)
        with pytest.raises(ValueError):
            B0(self.multiple_echoes_array, self.one_echo_list, self.affine)
        with pytest.raises(ValueError):
            B0(self.one_echo_array, self.multiple_echo_list, self.affine)
        with pytest.raises(ValueError):
            B0(self.multiple_echoes_array, self.multiple_echo_list,
               self.affine)

    def test_mask(self):
        # Create a mask where one of the echoes is True and the other is False
        mask = np.ones(self.correct_array.shape[:-1], dtype=bool)
        mask[:5, ...] = False

        all_pixels = B0(self.correct_array, self.correct_echo_list,
                        self.affine)
        masked_pixels = B0(self.correct_array, self.correct_echo_list,
                           self.affine, mask=mask)

        assert (all_pixels.phase_difference !=
                masked_pixels.phase_difference).any()
        assert (all_pixels.b0_map != masked_pixels.b0_map).any()
        assert (arraystats.ArrayStats(all_pixels.b0_map).calculate() !=
                arraystats.ArrayStats(masked_pixels.b0_map).calculate())

    def test_unwrap_phase(self):
        unwrapped = B0(self.correct_array, self.correct_echo_list, self.affine)
        wrapped = B0(self.correct_array, self.correct_echo_list, self.affine,
                     unwrap=False)

        assert (unwrapped.phase_difference != wrapped.phase_difference).any()
        assert (unwrapped.b0_map != wrapped.b0_map).any()
        assert (arraystats.ArrayStats(unwrapped.b0_map).calculate() !=
                arraystats.ArrayStats(wrapped.b0_map).calculate())

    def test_to_nifti(self):
        # Create a B0 map instance and test different export to NIFTI scenarios
        mapper = B0(self.correct_array, self.correct_echo_list,
                    self.affine, unwrap=False)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='b0test', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 5
        assert 'b0test_b0_map.nii.gz' in output_files
        assert 'b0test_mask.nii.gz' in output_files
        assert 'b0test_phase0.nii.gz' in output_files
        assert 'b0test_phase1.nii.gz' in output_files
        assert 'b0test_phase_difference.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='b0test', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only b0, phase0 and phase_difference are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='b0test', maps=['b0', 'phase0',
                                                       'phase_difference'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 3
        assert 'b0test_b0_map.nii.gz' in output_files
        assert 'b0test_phase0.nii.gz' in output_files
        assert 'b0test_phase_difference.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper = B0(self.correct_array, self.correct_echo_list,
                        self.affine, unwrap=False)
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='b0test', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')

    def test_pixel_array_type_assertion(self):
        # Empty array
        with pytest.raises(ValueError):
            mapper = B0(np.array([]), self.correct_echo_list, self.affine)
        # No input argument
        with pytest.raises(AttributeError):
            mapper = B0(None, self.correct_echo_list, self.affine)
        # List
        with pytest.raises(AttributeError):
            mapper = B0(list([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5]),
                        self.correct_echo_list, self.affine)
        # String
        with pytest.raises(AttributeError):
            mapper = B0("abcdef", self.correct_echo_list, self.affine)

    def test_echo_list_type_assertion(self):
        # Empty list
        with pytest.raises(ValueError):
            mapper = B0(self.correct_array, np.array([]), self.affine)
        # No input argument
        with pytest.raises(TypeError):
            mapper = B0(self.correct_array, None, self.affine)
        # Float
        with pytest.raises(TypeError):
            mapper = B0(self.correct_array, 3.2, self.affine)
        # String
        with pytest.raises(ValueError):
            mapper = B0(self.correct_array, "abcdef", self.affine)

    def test_real_data(self):
        # Get test data
        magnitude, phase, affine, te = fetch.b0_philips()
        te *= 1000

        # Process on a central slice only
        images = phase[:, :, 4, :]

        # Gold standard statistics
        gold_standard_b0 = [-34.174984, 189.285260, -1739.886907, 786.965213]

        # B0Map with unwrapping - Consider that unwrapping method may change
        mapper = B0(images, te, affine, unwrap=True)
        b0map_stats = arraystats.ArrayStats(mapper.b0_map).calculate()
        npt.assert_allclose([b0map_stats["mean"], b0map_stats["std"],
                            b0map_stats["min"], b0map_stats["max"]],
                            gold_standard_b0, rtol=0.01, atol=0)

    def test_b0_offset_correction(self):
        # Get test data that does not require b0_offset correction
        magnitude, phase, affine, te = fetch.b0_philips()
        te *= 1000
        # Process on a central slice only
        images = phase[:, :, 4, :]
        # B0Map with unwrapping
        mapper = B0(images, te, affine, unwrap=True)
        b0_map_without_offset_correction = (mapper.phase_difference /
                                            (2 * np.pi * mapper.delta_te))
        # This assertion proves that no offset correction was performed
        assert (mapper.b0_map == b0_map_without_offset_correction).all()

        # Get test data that requires b0_offset correction
        magnitude, phase, affine, te = fetch.b0_siemens(1)
        te *= 1000
        # Process on a central slice only
        images = phase[:, :, 4, :]
        # B0Map with unwrapping
        mapper = B0(images, te, affine, unwrap=True)
        b0_map_without_offset_correction = (mapper.phase_difference /
                                            (2 * np.pi * mapper.delta_te))
        # This assertion proves that there was offset correction performed
        assert (mapper.b0_map != b0_map_without_offset_correction).any()

# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import tempfile
import filecmp
from os import path
import pytest
from ukat.utils.siemens import write_dvs


class TestWriteDvs:
    # Initialise two different test schemes, one read from diffusion_scheme.txt
    # in the `test_data` directory and another initalised here as a string, str
    scheme_txt = path.abspath(path.join(path.dirname(__file__), "..", "tests",
                                        "test_data", "diffusion_scheme.txt"))
    scheme_str = (" 0.70710678          0.0   0.70710678      0\n"
                  " 0.70710678          0.0   0.70710678      5\n"
                  " 0.70710678          0.0   0.70710678    100\n"
                  " 0.70710678          0.0   0.70710678    500\n"
                  "-0.70710678   0.70710678          0.0      5\n"
                  "-0.70710678   0.70710678          0.0    100\n"
                  "-0.70710678   0.70710678          0.0    500")

    # Initialise the paths to the .dvs files in test_data whose contents are
    # the gold standard to compare against the contents of the .dvs files
    # generated in the tests below
    expected_from_txt = path.abspath(path.join(path.dirname(__file__), "..",
                                               "tests", "test_data",
                                               "expected_from_txt.dvs"))
    expected_from_str = path.abspath(path.join(path.dirname(__file__), "..",
                                               "tests", "test_data",
                                               "expected_from_str.dvs"))

    def test_from_txt(self):
        # Note that the calls to write_dvs are done within a context manager to
        # remove temporary files generated during the tests
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_filepath = path.join(tmpdir, "tmp")
            dvs_path, _ = write_dvs(self.scheme_txt, tmp_filepath,
                                    normalization='none',
                                    coordinate_system='xyz',
                                    comment='This is a comment')

            # Ensure output file (dvs_path) matches the expected in test_data
            assert filecmp.cmp(self.expected_from_txt, dvs_path)

    def test_from_str(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_filepath = path.join(tmpdir, "tmp")
            dvs_path, _ = write_dvs(self.scheme_str, tmp_filepath,
                                    normalization='none',
                                    coordinate_system='xyz',
                                    comment='This is a comment')

            # Ensure output file (dvs_path) matches the expected in test_data
            assert filecmp.cmp(self.expected_from_str, dvs_path)

    def test_bad_normalization(self):
        with pytest.raises(ValueError):
            write_dvs(self.scheme_str, "dummy_name", normalization='bad_norm')

    def test_bad_coordinate_system(self):
        with pytest.raises(ValueError):
            write_dvs(self.scheme_str, "dummy_name", coordinate_system='ijk')

    def test_bad_comment(self):
        with pytest.raises(ValueError):
            write_dvs(self.scheme_str, "dummy_name", comment=None)



import nibabel as nib
import numpy as np
import os

from sklearn.mixture import BayesianGaussianMixture


class Isnr:
    """
        Attributes
        ----------
        isnr : float
            Image signal to noise ratio.
        noise_mask : np.ndarray
            Mask of the ROI used to measure noise levels. Should be voxels
            outside the body.
        clusters : np.ndarray
            The labels assigned to each voxel during automatic background
            segmentation. This can be useful for debugging.
        """

    def __init__(self, pixel_array, noise_mask=None, n_clusters=3):
        """Initialise a image signal to noise ratio (iSNR) class instance.
        Parameters
        ----------
        pixel_array : np.ndarray
            Array of voxels over which iSNR should be calculated.
        noise_mask : np.ndarray, optional
            A binary voxel mask where voxels representing background i.e.
            outside the body, are True. If no mask is supplied, one is
            estimated using a Bayesian Gaussian mixture model to segment
            background voxels.
        n_clusters : int, optional
            When using the automatic background segmentation this is the
            total number of componenets the image is segmented into. The
            component with the lowest mean is assumed to be background.
            Default 3.
        """
        self.pixel_array = pixel_array
        self.n_clusters = n_clusters
        self.shape = pixel_array.shape
        self.dimensions = len(pixel_array.shape)

        if noise_mask is None:
            self.__mask_background__()
        else:
            self.noise_mask = noise_mask
            self.clusters = np.nan

        self.isnr = np.nan
        self.__snr__()

    def __mask_background__(self):
        np.random.seed(0)
        gmm = BayesianGaussianMixture(n_components=self.n_clusters,
                                      random_state=0,
                                      max_iter=500)
        # Because gmm's can get quite slow when fitting to large images,
        # we randomly sample a number of voxels equivalent to a 128 x 128 x
        # 3 image to keep runtimes consistent and manageable for large
        # multi-te/ti/dynamic images.
        fit_prop = (128 ** 2 * 3) / np.prod(self.shape)
        fit_mask = np.random.rand(self.pixel_array.size) < fit_prop
        gmm.fit(self.pixel_array.reshape(-1, 1)[fit_mask])
        self.clusters = gmm.predict(self.pixel_array.reshape(-1, 1)).reshape(
            self.shape)
        bg_label = np.argmin(gmm.means_)
        self.noise_mask = self.clusters == bg_label

    def __snr__(self):
        noise = np.std(self.pixel_array[self.noise_mask])
        signal = np.mean(self.pixel_array[~self.noise_mask])
        self.isnr = (signal/noise) * np.sqrt(2-(np.pi/2))


class Tsnr:
    """
    Attributes
    ----------
    tsnr_map : np.ndarray
        Map of temporal signal to noise ratio.
    """

    def __init__(self, pixel_array, affine, mask=None):
        """Initialise a temporal signal to noise ratio (tSNR) class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            A array containing the signal from each voxel with the last
            dimension being repeated dynamics i.e. the array needed to
            generate a tSNR map would have dimensions [x, y, z, d].
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired tSNR map rather than the raw data i.e. omit the dynamics
            dimension.
        """
        np.seterr(divide='ignore', invalid='ignore')
        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.dimensions = len(pixel_array.shape)
        self.n_d = pixel_array.shape[-1]
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
        # Don't process any nan values
        self.mask[np.isnan(np.sum(pixel_array, axis=-1))] = False

        self.pixel_array = self.pixel_array * \
            np.repeat(self.mask[..., np.newaxis],
                      self.n_d, axis=-1)

        # Initialise output attributes
        self.tsnr_map = np.zeros(self.shape)
        self.tsnr_map = self.__tsnr__()

    def __tsnr__(self):
        # Regress out linear and quadratic temporal drifts associated with
        # hardware using a GLM of the form Y = X * beta + error
        # as in Hutton C et al. The impact of physiological noise correction
        # on fMRI at 7T. NeuroImage 2011;57:101–112 doi:
        # 10.1016/j.neuroimage.2011.04.018.

        # Vectorise image
        pixel_array_vector = np.reshape(self.pixel_array,
                                        (np.prod(self.shape), self.n_d))
        x = np.vstack([np.ones(self.n_d),
                       np.arange(1, self.n_d + 1),
                       np.arange(1, self.n_d + 1) ** 2]).T
        beta = np.linalg.pinv(x).dot(pixel_array_vector.T)
        pixel_array_vector_detrended = pixel_array_vector.T - \
            x[:, 1:].dot(beta[1:])
        pixel_array_detrended = pixel_array_vector_detrended.T.reshape((
            *self.shape, self.n_d))
        tsnr_map = pixel_array_detrended.mean(axis=-1) / \
            pixel_array_detrended.std(axis=-1)  # Might want to try
        # ddof=1 as per Kevins code...
        tsnr_map[tsnr_map > 1000] = 0
        tsnr_map = np.nan_to_num(tsnr_map)
        return tsnr_map

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output'):
        """Exports tSNR maps to NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)

        tsnr_nifti = nib.Nifti1Image(self.tsnr_map, affine=self.affine)
        nib.save(tsnr_nifti, base_path + '_tsnr_map.nii.gz')



import numpy as np
import pytest
from ukat.utils.ge import scale_b1


class TestScaleB1:
    unscalled_b1 = np.ones((10, 10, 5)) * 200

    def test_scale_b1(self):
        scaled_b1 = scale_b1(self.unscalled_b1, 20)
        assert np.allclose(scaled_b1, 100)
